from corems.encapsulation.factory.processingSetting  import MolecularFormulaSearchSettings, TransientSetting, MassSpecPeakSetting, MassSpectrumSetting
from corems.encapsulation.factory.processingSetting  import CompoundSearchSettings, GasChromatographSetting
from corems.encapsulation.factory.processingSetting  import DataInputSetting

class MSParameters:

    molecular_search =  MolecularFormulaSearchSettings()
    transient =  TransientSetting()
    mass_spectrum = MassSpectrumSetting()
    ms_peak =  MassSpecPeakSetting()
    data_input =  DataInputSetting()
  
class GCMSParameters:

    molecular_search = CompoundSearchSettings()
    gc_ms = GasChromatographSetting()
    

def default_parameters(file_location): #pragma: no cover

        parameters = dict()

        parameters["Aterm"] = 0

        parameters["Bterm"] = 0

        parameters["Cterm"] = 0

        parameters["exc_high_freq"] = 0

        parameters["exc_low_freq"] = 0

        parameters["bandwidth"] = 0

        parameters['analyzer'] = 'Unknown'
        
        parameters['instrument_label'] = 'Unknown' 

        parameters['sample_name'] = 'Unknown'

        parameters["number_data_points"] = 0

        parameters["polarity"] = 'Unknown'

        parameters["filename_path"] = str(file_location)

        """scan_number and rt will be need to lc ms"""

        parameters["mobility_scan"] = 0

        parameters["mobility_rt"] = 0

        parameters["scan_number"] = 0

        parameters["rt"] = 0

        return parameters    


import json

from pymongo import MongoClient
from corems.encapsulation.factory.parameters import MSParameters

class MolForm_Mongo:
    
    '''this class is outdated
    'exists only for reference
    '''
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        # make sure the dbconnection gets closed
        self.client.close()

    def __enter__(self):
        
        self.client = MongoClient("mongodb://corems-client:esmlpnnl2019@localhost:27017/corems")
        
        db = self.client.corems
        self.molform_collection = db.molform
        return self
       
    def add_all(self, dic_molform_list):
        
        try:
            self.molform_collection.insert_many(dic_molform_list, ordered=False)
        except Exception as e:
            #duplicate_error
            try:
                if e.details["writeErrors"][0]["code"] == 11000:
                    print('ohh nooo')
            except:
                raise  Exception(e)        
    
    def check_entry(self,classe, ion_type):
       
        formulas = self.molform_collection.find({'classe': classe, 'ion_type': ion_type}).limit(1)
        
        if list(formulas):
            return True
        else:
            return False    
        
    def add_entry(self, dic_molform): 

        try:
            self.molform_collection.insert(dic_molform)
        except Exception as e:
            #duplicate_error
            try:
                if e.details["writeErrors"][0]["code"] == 11000:
                    pass
            except:
                raise  Exception(e)   
  
    def get_all(self):

        return self.molform_collection.find({})

    def get_dict_entries(self, classe, ion_type, nominal_mzs):

        dict_res = {}

        formulas = self.molform_collection.find( {'classe': {"$in": classe}, 
                                                'ion_type': ion_type,
                                                'nominal_mz':{"$in": nominal_mzs},  
                                                'O_C' : { '$lt': MSParameters.molecular_search.oc_filter }, 
                                                'H_C' : { '$gt': MSParameters.molecular_search.min_hc_filter},
                                                'DBE' : { '$gt': MSParameters.molecular_search.min_dbe},
                                                'DBE' : { '$lt': MSParameters.molecular_search.max_dbe},
                                                })
        for formula in formulas:
            
            if formula["classe"] in dict_res.keys():
                
                if formula['nominal_mz'] in dict_res[formula['classe']].keys():
                    
                    dict_res.get(formula['classe']).get(formula['nominal_mz']).append(json.loads(formula['mol_formula']) )
                
                else:

                    dict_res.get(formula['classe'])[formula['nominal_mz']] = [json.loads(formula['mol_formula']) ]  
        
            else:
                
                dict_res[formula['classe']] = {formula['nominal_mz']: [json.loads(formula['mol_formula'])] }     
        
        return dict_res
    
    def get_entries(self,classe, ion_type, nominal_mz):
        
        #print (classe, ion_type, nominal_mz)
        
        formulas = self.molform_collection.find( {'classe': classe, 
                                                'ion_type': ion_type,
                                                'nominal_mz': nominal_mz,  
                                                'O_C' : { '$lt': MSParameters.molecular_search.oc_filter }, 
                                                'H_C' : { '$gt': MSParameters.molecular_search.min_hc_filter},
                                                'DBE' : { '$gt': MSParameters.molecular_search.min_dbe},
                                                'DBE' : { '$lt': MSParameters.molecular_search.max_dbe},
                                                })
        

        return [json.loads(formula['mol_formula']) for formula in formulas]
        
   
    def update_entry(self, entry):
        
        raise NotImplementedError 

    def delete_entry(self, entry):
        
        raise NotImplementedError 



from pathlib import Path
import json

from corems.encapsulation.factory.processingSetting  import MolecularFormulaSearchSettings, TransientSetting
from corems.encapsulation.factory.processingSetting  import MassSpectrumSetting
from corems.encapsulation.factory.processingSetting  import MassSpecPeakSetting
from corems.encapsulation.factory.processingSetting  import GasChromatographSetting
from corems.encapsulation.factory.processingSetting import CompoundSearchSettings, DataInputSetting

def load_and_set_parameters_ms(mass_spec_obj, parameters_path=False):   
    
    if parameters_path:
        
        file_path = Path(parameters_path)

    else:
        
        filename='SettingsCoreMS.json'
        file_path = Path.cwd() / filename 

    if file_path.exists():  

            with open(file_path, 'r', encoding='utf8',) as stream:
                data_loaded = json.load(stream)
                _set_dict_data_ms(data_loaded, mass_spec_obj)
    else:
        
        raise FileNotFoundError("Could not locate %s", file_path)   

def load_and_set_parameters_gcms(gcms_obj, parameters_path=False):   
    
    if parameters_path:
        
        file_path = Path(parameters_path)

    else:
        
        filename='SettingsCoreMS.json'
        file_path = Path.cwd() / filename 

    if file_path.exists():  

            with open(file_path, 'r', encoding='utf8',) as stream:
                data_loaded = json.load(stream)
                _set_dict_data_gcms(data_loaded, gcms_obj)
    else:
        
        raise FileNotFoundError("Could not locate %s", file_path)   
    
def _set_dict_data_gcms(data_loaded, gcms_obj):
    
    classes = [GasChromatographSetting(),
               CompoundSearchSettings(),
              ]

    labels = ["GasChromatograph", "MolecularSearch"]
    
    label_class = zip(labels, classes)

    if data_loaded:
    
        for label, classe in label_class:
            class_data = data_loaded.get(label)
            # not always we will not all the settings
            # this allow a class data to be none and continue
            # to import the other classes
            if class_data:
                for item, value in class_data.items():
                    setattr(classe, item, value)

    gcms_obj.chromatogram_settings = classes[0]
    gcms_obj.molecular_search_settings = classes[1]

def _set_dict_data_ms(data_loaded, mass_spec_obj):
    
    from copy import deepcopy

    classes = [MolecularFormulaSearchSettings(), 
               TransientSetting(),
               MassSpectrumSetting(),
               MassSpecPeakSetting()
               ]
               
    labels = ["MolecularFormulaSearch", "Transient", "MassSpectrum", "MassSpecPeak"]
    
    label_class = zip(labels, classes)

    if data_loaded:
    
        for label, classe in label_class:
            class_data = data_loaded.get(label)
            # not always we will have all the settings classes
            # this allow a class data to be none and continue
            # to import the other classes
            if class_data:
                for item, value in class_data.items():
                    setattr(classe, item, value)
    
    mass_spec_obj.molecular_search_settings = classes[0]
    mass_spec_obj.transient_settings = classes[1]
    mass_spec_obj.settings = classes[2]
    mass_spec_obj.mspeaks_settings = classes[3]


def load_and_set_parameters_class(parameter_label, instance_parameters_class, parameters_path=False):   
    
    if parameters_path: file_path = Path(parameters_path)

    else: file_path = Path.cwd() / 'SettingsCoreMS.json' 
        
    if file_path.exists():
        
        with open(file_path, 'r', encoding='utf8',) as stream:
            
            data_loaded = json.load(stream)
            parameter_class = _set_dict_data(data_loaded, parameter_label, instance_parameters_class)
            
            return parameter_class
    else:
        
        raise FileNotFoundError("Could not locate %s", file_path)  
    
def _set_dict_data(data_loaded, parameter_label, instance_ParameterClass):
    
    classes = [instance_ParameterClass]
               
    labels = [parameter_label]
    
    label_class = zip(labels, classes)

    if data_loaded:
    
        for label, classe in label_class:
            class_data = data_loaded.get(label)
            # not always we will have all the settings classes
            # this allow a class data to be none and continue
            # to import the other classes
            if class_data:
                for item, value in class_data.items():
                    setattr(classe, item, value)
    
    return classes[0]


import numpy as np
import scipy.stats
'''exploratory module based on  Yuanyue Li code at 
TODO add GitHub and Paper here'''

def entropy_distance(v, y):
    merged = v + y
    entropy_increase = 2 * scipy.stats.entropy(merged) - scipy.stats.entropy(v) - scipy.stats.entropy(y)
    return entropy_increase

def _weight_intensity_for_entropy(x):

    if sum(x) > 0:
        WEIGHT_START = 0.25
        WEIGHT_SLOPE = 0.5

        entropy_x = scipy.stats.entropy(x)
        weight = WEIGHT_START + WEIGHT_SLOPE * entropy_x
        x = np.power(x, weight)
        x = x / sum(x)
        return x


def weighted_entropy_distance(v, y):
    v = _weight_intensity_for_entropy(v)
    y = _weight_intensity_for_entropy(y)

    merged = v + y
    entropy_increase = 2 * scipy.stats.entropy(merged) - scipy.stats.entropy(v) - scipy.stats.entropy(y)
    return entropy_increase


def chebyshev_distance(v, y):
    r"""
    Chebyshev distance:

    .. math::

        \underset{i}{\max}{(|v_{i}\ -\ y_{i}|)}
    """
    return np.max(np.abs(v - y))


def squared_euclidean_distance(v, y):
    r"""
    Squared Euclidean distance:

    .. math::

        \sum(v_{i}-y_{i})^2
    """
    return np.sum(np.power(v - y, 2))


def fidelity_similarity(v, y):
    r"""
    Fidelity similarity:

    .. math::

        \sum\sqrt{v_{i}y_{i}}
    """
    return np.sum(np.sqrt(v * y))


def matusita_distance(v, y):
    r"""
    Matusita distance:

    .. math::

        \sqrt{\sum(\sqrt{v_{i}}-\sqrt{y_{i}})^2}
    """
    return np.sqrt(np.sum(np.power(np.sqrt(v) - np.sqrt(y), 2)))


def squared_chord_distance(v, y):
    r"""
    Squared-chord distance:

    .. math::

        \sum(\sqrt{v_{i}}-\sqrt{y_{i}})^2
    """
    return np.sum(np.power(np.sqrt(v) - np.sqrt(y), 2))


def bhattacharya_1_distance(v, y):
    r"""
    Bhattacharya 1 distance:

    .. math::

        (\arccos{(\sum\sqrt{v_{i}y_{i}})})^2
    """
    s = np.sum(np.sqrt(v * y))
    # TODO:Fix this!
    if s > 1:
        if s > 1 + 1e-6:
            print("Error in calculating Bhattacharya 1 distance, got arccos {}".format(s))
        s = 1
    return np.power(np.arccos(s), 2)


def bhattacharya_2_distance(v, y):
    r"""
    Bhattacharya 2 distance:

    .. math::

        -\ln{(\sum\sqrt{v_{i}y_{i}})}
    """
    s = np.sum(np.sqrt(v * y))
    if s == 0:
        return np.inf
    else:
        return -np.log(s)


def harmonic_mean_similarity(v, y):
    r"""
    Harmonic mean similarity:

    .. math::

        #1-2\sum(\frac{v_{i}y_{i}}{v_{i}+y_{i}})
        2\sum(\frac{v_{i}y_{i}}{v_{i}+y_{i}})
    """
    #return 1 - 2 * np.sum(v * y / (v + y))
    return 2 * np.sum(v * y / (v + y))


#def pearson_chi_squared_distance(v, y):
#    r"""
#    Pearson χ2 distance:
#
#    .. math::
#
#        \sum\frac{(v_{i}-y_{i})^2}{y_{i}}
#    """
#    return np.sum(np.power(v - y, 2) / y)


#def neyman_chi_squared_distance(v, y):
#    r"""
#    Neyman χ2 distance:
#
#    .. math::
#
#        \sum\frac{(v_{i}-y_{i})^2}{v_{i}}
#    """
#    return np.sum(np.power(v - y, 2) / v)


#def probabilistic_symmetric_chi_squared_distance(v, y):
#    r"""
#    Probabilistic symmetric χ2 distance:
#
#    .. math::
#
#        \frac{1}{2} \times \sum\frac{(v_{i}-y_{i}\ )^2}{v_{i}+y_{i}\ }
#    """
#    return 1 / 2 * np.sum(np.power(v - y, 2) / (v + y))


#def topsoe_distance(v, y):
#    r"""
#    Topsøe distance:
#
#    .. math::
#
#        \sum{(v_{i}ln\frac{v_{i}}{Z_i}+y_{i}ln\frac{y_{i}}{Z_i}),\ \ \ Z_i=\frac{1}{2}(v_{i}+y_{i})}
#    """
#    z = 1 / 2 * (v + y)
#    z[z == 0] = 1
#    vz = v / z
#    yz = y / z
#    vz[v == 0] = 1
#    yz[y == 0] = 1
#    return np.sum(v * np.log(vz) + y * np.log(yz))


def chernoff_distance(v, y):
    r"""
    Chernoff distance:

    .. math::

        \max{(-ln\sum(v_{i}^ty_{i}^{1-t})^{1-t})},\ t=0.1,\ 0\le\ t<1
    """
    t = 0.1
    return np.max(-np.log(
        np.sum(np.power(np.power(v, t) * np.power(y, 1 - t), 1 - t))))


def ruzicka_distance(v, y):
    r"""
    Ruzicka distance:

    .. math::

        \frac{\sum{|v_{i}-y_{i}|}}{\sum{\max(v_{i},y_{i})}}
    """
    dist = np.sum(np.abs(v - y)) / np.sum(np.maximum(v, y))
    return dist


def roberts_distance(v, y):
    r"""
    Roberts distance:

    .. math::

        1-\sum\frac{(v_{i}+y_{i})\frac{\min{(v_{i},y_{i})}}{\max{(v_{i},y_{i})}}}{\sum(v_{i}+y_{i})}
    """
    return 1 - np.sum((v + y) * np.minimum(v, y) / np.maximum(v, y) / np.sum(v + y))


def intersection_distance(v, y):
    r"""
    Intersection distance:

    .. math::

        1-\frac{\sum\min{(v_{i},y_{i})}}{\min(\sum{v_{i},\sum{y_{i})}}}
    """
    return 1 - np.sum(np.minimum(v, y)) / min(np.sum(v), np.sum(y))


def motyka_distance(v, y):
    r"""
    Motyka distance:

    .. math::

        -\frac{\sum\min{(y_{i},v_{i})}}{\sum(y_{i}+v_{i})}
    """
    dist = np.sum(np.minimum(v, y)) / np.sum(v + y)
    return dist


def canberra_distance(v, y):
    r"""
    Canberra distance:

    .. math::

        #\sum\frac{|v_{i}-y_{i}|}{|v_{i}|+|y_{i}|}
        \sum_{i}\frac{|y_{i} - v_{i}|}{y_{i} + v_{i}}
    """
    #return np.sum(np.abs(v - y) / (np.abs(v) + np.abs(y)))
    return np.sum(np.abs(y - v)/(y + v))

def canberra_metric(v, y):
    r"""
    Canberra Metric

    .. math::

        \frac{1}{\sum_{i}I(v_{i}\neq 0)}\sum_{i}\frac{|y_{i}-v_{i}|}{(y_{i}+v_{i})}
    """

    return (1 / np.sum(v > 0)) * np.sum(np.abs(y - v)/(y + v))


def kulczynski_1_distance(v, y):
    r"""
    Kulczynski 1 distance:

    .. math::

        \frac{\sum{|v_i}-y_i|}{\sum m\ i\ n\ (v_i,y_i)}
    """
    return np.sum(np.abs(y - v)) / np.sum(np.minimum(y, v))


def baroni_urbani_buser_distance(v, y):
    r"""
    Baroni-Urbani-Buser distance:

    .. math::

        1-\frac{\sum\min{(v_i,y_i)}+\sqrt{\sum\min{(v_i,y_i)}\sum(\max{(v)}-\max{(v_i,y_i)})}}{\sum{\max{(v_i,y_i)}+\sqrt{\sum{\min{(v_i,y_i)}\sum(\max{(v)}-\max{(v_i,y_i)})}}}}
    """
    if np.max(v) < np.max(y):
        v, y = y, v
    d1 = np.sqrt(np.sum(np.minimum(v, y) * np.sum(max(v) - np.maximum(v, y))))
    return 1 - (np.sum(np.minimum(v, y)) + d1) / (np.sum(np.maximum(v, y)) + d1)


def penrose_size_distance(v, y):
    r"""
    Penrose size distance:

    .. math::

        \sqrt N\sum{|y_i-v_i|}
    """
    n = np.sum(v > 0)
    return np.sqrt(n) * np.sum(np.abs(y - v))


def mean_character_distance(v, y):
    r"""
    Mean character distance:

    .. math::

        \frac{1}{N}\sum{|y_i-v_i|}
    """
    n = np.sum(v > 0)
    return 1 / n * np.sum(np.abs(y - v))


def lorentzian_distance(v, y):
    r"""
    Lorentzian distance:

    .. math::

        \sum{\ln(1+|v_i-y_i|)}
    """
    return np.sum(np.log(1 + np.abs(y - v)))


def penrose_shape_distance(v, y):
    r"""
    Penrose shape distance:

    .. math::

        \sqrt{\sum((v_i-\bar{v})-(y_i-\bar{y}))^2}
    """
    v_avg = np.mean(v)
    y_avg = np.mean(y)
    return np.sqrt(np.sum(np.power((y - y_avg) - (v - v_avg), 2)))


def clark_distance(v, y):
    r"""
    Clark distance:

    .. math::

        #(\frac{1}{N}\sum(\frac{v_i-y_i}{|v_i|+|y_i|})^2)^\frac{1}{2}
        \sqrt{\sum(\frac{|v_i-y_i|}{v_i+y_i})^2}
    """
    #n = np.sum(v > 0)
    #return np.sqrt(1 / n * np.sum(np.power((v - y) / (np.abs(v) + np.abs(y)), 2)))
    return np.sqrt(np.sum(np.power(np.abs(y - v) / (y + v), 2)))


def hellinger_distance(v, y):
    r"""
    Hellinger distance:

    .. math::

        #\sqrt{2\sum(\sqrt{\frac{v_i}{\bar{v}}}-\sqrt{\frac{y_i}{\bar{y}}})^2}
        \sqrt{2\sum(\sqrt{v_i}-\sqrt{y_i})^2}
    """
    #v_avg = np.mean(v)
    #y_avg = np.mean(y)
    #return np.sqrt(2 * np.sum(np.power(np.sqrt(v / v_avg) - np.sqrt(y / y_avg), 2)))
    return np.sqrt(2 * np.sum(np.power(np.sqrt(y) - np.sqrt(v), 2)))


def whittaker_index_of_association_distance(v, y):
    r"""
    Whittaker index of association distance:

    .. math::

        \frac{1}{2}\sum|\frac{v_i}{\bar{v}}-\frac{y_i}{\bar{y}}|
    """
    v_avg = np.mean(v)
    y_avg = np.mean(y)
    return 1 / 2 * np.sum(np.abs(v / v_avg - y / y_avg))


#def symmetric_chi_squared_distance(v, y):
#    r"""
#    Symmetric χ2 distance:
#
#    .. math::
#
#        \sqrt{\sum{\frac{\bar{v}+\bar{y}}{N(\bar{v}+\bar{y})^2}\frac{(v_i\bar{y}-y_i\bar{v})^2}{v_i+y_i}\ }}
#    """
#    v_avg = np.mean(v)
#    y_avg = np.mean(y)
#    n = np.sum(v > 0)
#
#    d1 = (v_avg + y_avg) / (n * np.power(v_avg + y_avg, 2))
#    return np.sqrt(d1 * np.sum(np.power(v * y_avg - y * v_avg, 2) / (v + y)))

def similarity_index_distance(v, y):
    r"""
    Similarity Index Distance:

    .. math::

        \sqrt{\frac{\sum\{\frac{v_i-y_i}{y_i}\}^2}{N}}
    """
    n = np.sum(v > 0)
    return np.sqrt(1 / n * np.sum(np.power((v - y) / y, 2)))


def improved_similarity_distance(v, y):
    r"""
    Improved Similarity Index:

    .. math::

        \sqrt{\frac{1}{N}\sum\{\frac{y_i-v_i}{y_i+v_i}\}^2}
    """
    n = np.sum(v > 0)
    return np.sqrt(1 / n * np.sum(np.power((y - v) / (y + v), 2)))


def absolute_value_distance(v, y):
    r"""
    Absolute Value Distance:

    .. math::

        \frac { \sum(|y_i-v_i|)}{\sum v_i}

    """
    dist = np.sum(np.abs(y - v)) / np.sum(v)
    return dist

def spectral_contrast_angle_distance(v, y):
    r"""
    Spectral Contrast Angle:

    .. math::

        1 - \frac{\sum{y_iv_i}}{\sqrt{\sum y_i^2\sum v_i^2}}
        \arccos(\frac{\sum_{P}y_{p}^* v_{p}^*}{\sqrt{\sum_{P}y_{p}^{*2} \sum_{P}v_{p}^{*2}}})
    """
    #return 1 - np.sum(y * v) / \
    #       np.sqrt(np.sum(np.power(y, 2)) * np.sum(np.power(v, 2)))

    return np.arccos(np.sum(y * v) / (np.sqrt(np.sum(np.power(y, 2)) * np.sum(np.power(v, 2)))))


def wave_hedges_distance(v, y):
    r"""
    Wave Hedges distance:

    .. math::

        \sum\frac{|v_i-y_i|}{\max{(v_i,y_i)}}
    """
    return np.sum(np.abs(v - y) / np.maximum(v, y))

def dice_similarity(v, y):
    r"""
    Dice similarity:

    .. math::

        \frac{\sum(v_i-y_i)^2}{\sum v_i^2+\sum y_i^2}
        \frac{2 * \sum_{i}v_{i}y_{i}}{\sum_{i}y_{i}^2 + \sum_{i}v_{i}^2}
    """
    return 2 * np.sum(v * y) / (np.sum(np.power(v, 2)) + np.sum(np.power(y, 2)))


def inner_product_distance(v, y):
    r"""
    Inner Product distance:

    .. math::

        1-\sum{v_iy_i}
    """
    return 1 - np.sum(v * y)


def divergence_distance(v, y):
    r"""
    Divergence distance:

    .. math::

        2\sum\frac{(v_i-y_i)^2}{(v_i+y_i)^2}
    """
    return 2 * np.sum((np.power(v - y, 2)) / np.power(v + y, 2))


def _chi_squared_distance(v, y):
    r"""
    Additive symmetric χ2 distance:

    .. math::

        \sum\frac{(v_i-y_i)^2(v_i+y_i)}{v_iy_i}
    """
    dist = np.sum(np.power(v - y, 2) * (v + y) / (v * y))
    return dist


def jensen_difference_distance(v, y):
    r"""
    Jensen difference:

    .. math::

        \sum[\frac{1}{2}(v_i\ln{v_i}+y_i\ln{y_i})-(\frac{v_i+y_i}{2})\ln{(\frac{v_i+y_i}{2})}]
    """
    y_v_avg = (y + v) / 2
    return np.sum(
        1 / 2 * (y * np.log(y) + v * np.log(v)) -
        y_v_avg * np.log(y_v_avg)
    )


def kumar_johnson_distance(v, y):
    r"""
    Kumar-Johnson distance:

    .. math::

        \sum\frac{(v_i^2-y_i^2)^2}{2(v_iy_i)^\frac{3}{2}}
    """
    return np.sum(
        np.power(np.power(v, 2) - np.power(y, 2), 2) / \
        (2 * np.power(v * y, 3 / 2))
    )


def avg_l_distance(v, y):
    r"""
    Avg (L1, L∞) distance:

    .. math::

        \frac{1}{2}(\sum|v_i-y_i|+\underset{i}{\max}{|v_i-y_i|})
    """
    return 1 / 2 * (np.sum(np.abs(v - y)) + max(np.abs(v - y)))


def vicis_wave_hadges_distance(v, y):
    r"""
    Vicis-Wave Hadges distance:

    .. math::

        \sum\frac{|v_i-y_i|}{\min{(v_i,\ y_i)}}
    """
    return np.sum(np.abs(v - y) / np.minimum(v, y))


def vicis_symmetric_chi_squared_1_distance(v, y):
    r"""
    Vicis-Symmetric χ2 1 distance:

    .. math::

        \sum\frac{(v_i-y_i)^2}{\min{(v_i,y_i)^2}}
    """
    return np.sum(np.power(v - y, 2) / np.power(np.minimum(v, y), 2))


def vicis_symmetric_chi_squared_2_distance(v, y):
    r"""
    Vicis-Symmetric χ2 2 distance:

    .. math::

        \sum\frac{(v_i-y_i)^2}{\min{(v_i,y_i)}}
    """
    return np.sum(np.power(v - y, 2) / np.minimum(v, y))


def vicis_symmetric_chi_squared_3_distance(v, y):
    r"""
    Vicis-Symmetric χ2 3 distance:

    .. math::

        \sum\frac{(v_i-y_i)^2}{\max{(v_i,y_i)}}
    """
    return np.sum(np.power(v - y, 2) / np.maximum(v, y))


def max_symmetric_chi_squared_distance(v, y):
    r"""
    Max-Symmetric χ2 distance:

    .. math::

        \max{(\sum\frac{(v_i-y_i)^2}{v_i},\sum\frac{(v_i-y_i)^2}{y_i})}
    """
    return max(np.sum(np.power(v - y, 2) / v), np.sum(np.power(v - y, 2) / y))


def min_symmetric_chi_squared_distance(v, y):
    r"""
    Min-Symmetric χ2 distance:

    .. math::

        \min{(\sum\frac{(v_i-y_i)^2}{v_i},\sum\frac{(v_i-y_i)^2}{y_i})}
    """
    return min(np.sum(np.power(v - y, 2) / v), np.sum(np.power(v - y, 2) / y))

"""added by Allison"""
def additive_sym_chi_sq(v, y):
    r"""
    Additive Symmetric χ2 distance:

    .. math::

        \sum_{i}\frac{(y_{i} - v_{i})^2(y_{i}+v_{i})}{y_{i}v_{i}}
    """
    return np.sum((np.power(y - v, 2) * (y + v))/(y * v))

def bhattacharya_distance(v, y):
    r"""
    Bhattacharya Distance:

    .. math::

        -ln(\sum_{i}\sqrt{y_{i}v_{i}})
    """
    return -1 * np.log(np.sum(np.sqrt(y * v)))

def generalized_ochai_index(v, y):
    r"""
    Generalized Ochai Index

    .. math::

        1 - \frac{\sum_{i}min(y_{i}, v_{i})}{\sqrt{\sum_{i}y_{i} \sum_{i}v_{i}}}
    """

    ind = np.sum(np.minimum(y, v)) / np.sqrt(np.sum(y) * np.sum(v))
    return 1 - ind 

def gower_distance(v, y):
    r"""
    Gower Distance

    .. math::

        \frac{1}{N}\sum_{i}|y_{i} - v_{i}|
    """

    n = np.sum(y > 0)
    return (1 / n) * np.sum(np.abs(y - v))

def impr_sqrt_cosine_sim(v, y):
    r"""
    Improved Square Root Cosine Similarity

    .. math::

        \frac{\sum_{i}\sqrt{y_{i}v_{i}}}{\sum_{i}\sqrt{y_{i}}\sum_{i}\sqrt{v_{i}}}
    """

    return np.sum(np.sqrt(y * v)) / (np.sum(np.sqrt(y)) * np.sum(np.sqrt(v)))

def intersection_sim(v, y):
    r"""
    Intersection Similarity

    .. math::

        \sum_{i}min(y_{i}, v_{i})
    """

    return np.sum(np.minimum(y, v))

def j_divergence(v, y):
    r"""
    J Divergence

    .. math::
        
        \sum_{i}(y_{i} - v_{i}) ln(\frac{y_{i}}{v_{i}})
    """

    return np.sum((v - y) * np.log(v / y))

def jensen_shannon_index(v, y):
    r"""
    Jensen-Shannon Index

    .. math::

        \frac{1}{2}[\sum_{i}y_{i}ln(\frac{2y_{i}}{y_{i} + v_{i}}) + \sum_{i}v_{i}ln(\frac{2v_{i}}{y_{i}+v_{i}})]
    """

    return (1 / 2) * (np.sum(y * np.log(2 * y / (y + v))) + np.sum(v * np.log(2 * v / (y + v))))

def k_divergence(v, y):
    r"""
    K-Divergence

    .. math::

        \sum_{i}y_{i}ln(\frac{2y_{i}}{y_{i} + v_{i}})
    """

    return np.sum(v * np.log((2 * v) / (y + v)))

"""added by Chae"""
def topsoe_distance(v, y):
    r""" Fixed
    "I commented out the previous one; please review"
    """
    #v[v==0] = 1 #added by amt
    #y[y==0] = 1 #added by amt
    return np.sum((y * np.log((2 * y)/(y + v))) + (v * np.log((2 * v)/(y + v))))

def probabilistic_symmetric_chi_squared_distance(v, y):
    r""" Fixed
    "I commented out the previous one; please review"
    """
    return 2 * np.sum(np.sum(np.power(y - v, 2) / (y + v)))

def VW6(v, y):
    r"""
    "appears to be the same as max_symmetric_chi_squared_distance"
    """
    return min(np.sum(np.power(y - v, 2) / y), np.sum(np.power(y - v, 2) / v))

def VW5(v, y):
    r"""
    "appears to be the same as max_symmetric_chi_squared_distance"
    """
    return max(np.sum(np.power(y - v, 2) / y), np.sum(np.power(y - v, 2) / v))

def VW4(v, y):
    r"""
    "Tecnically the Symmetric chi2 eq63"
    """
    return np.sum(np.power(y - v, 2) / np.maximum(y, v))

def VW3(v, y):
    r"""
    "New"
    """
    return np.sum(np.power(y - v, 2) / np.minimum(y, v))

def VW2(v, y):
    r"""
    "New"
    """
    return np.sum(np.power(y - v, 2) / np.power(np.minimum(y, v), 2))

def VW1(v, y):
    r"""
    "New"
    """
    return np.sum(np.abs(y - v) / np.minimum(y, v))

def taneja_divergence(v, y):
    r"""
    "New"
    """
    return np.sum(((y + v) / 2) * np.log((y + v)/(2 * np.sqrt(y * v))))

def symmetric_chi_squared_distance (v, y):
    r"""
    "New"
    """
    return np.sum(np.power(y - v, 2) / (y * v))

def squared_chi_squared_distance(v, y):
    r"""
    "New"
    """
    return np.sum(np.power(y - v, 2) / (y + v))

def square_root_cosine_correlation(v, y):
    r"""
    "New"
    """
    return np.sum(np.sqrt(y * v)) / (np.sum(y) * np.sum(v))

def sorensen_distance(v, y):
    r"""
    "New"
    """
    return np.sum(np.abs(y - v)) / (np.sum(y + v))

def Pearson_chi_squared_distance(v, y):
    r"""
    "New"
    """
    return np.sum(np.power(y - v, 2) / v)

def Neyman_chi_squared_distance(v, y):
    r"""
    "New"
    """
    return np.sum(np.power(y - v, 2) / y)

def Minokowski_3(v, y):
    r"""
    "New"
    """
    return np.power(np.sum(np.power(np.abs(y - v), 3)), 1/3)

def Minokowski_4(v, y):
    r"""
    "New"
    """
    return np.power(np.sum(np.power(np.abs(y - v), 4)), 1/4)

def kumarjohnson_divergence(v, y):
    r"""
    "New"
    """
    return np.sum(np.power(np.power(y, 2) + np.power(v, 2), 2) / (2* np.power(y * v, 3/2)))

def kumarhassebrook_similarity(v, y):
    r"""
    "New"
    """
    return np.sum(y * v) / (np.sum(np.power(y, 2)) + np.sum(np.power(v, 2)) - np.sum(y * v))

def kullbackleibler_divergence (v, y):
    r"""
    "New"
    """
    return np.sum(v * np.log(v / y))

def soergel_distance(v, y):
    r"""
    "New"
    """
    return np.sum(np.abs(y - v))/np.sum(np.maximum(y, v))



__author__ = "Yuri E. Corilo"
__date__ = "Feb 12, 2020"

from pathlib import Path
from threading import Thread
from io import BytesIO

from netCDF4 import Dataset
from s3path import S3Path

from corems.encapsulation.constant import Labels
from corems.encapsulation.factory.parameters import default_parameters
from corems.mass_spectra.factory.GC_Class import GCMSBase
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecCentroidLowRes


class ReadAndiNetCDF(Thread):

	def __init__(self, file_location, analyzer='Quadruple', instrument_label='GCMS-Agilent', auto_process=True):
		'''
		 # Parameters
		----------
		## file_location : Path or S3Path
        file_location is full data path
		'''

		Thread.__init__(self)
		
		if  isinstance(file_location, str):
			# if obj is a string it defaults to create a Path obj, pass the S3Path if needed
			self.file_location = Path(file_location)
		
		else:
			
			self.file_location = file_location

		if not self.file_location.exists():

			raise FileNotFoundError("File does not exist at %s", file_location)

		if isinstance(file_location, S3Path):
			bytes_io = self.file_location.open('rb').read()
		
			self.net_cdf_obj = Dataset(self.file_location.name, "r", diskless=True, memory=bytes_io, format='NETCDF3_CLASSIC')
		
		else:
			self.net_cdf_obj = Dataset(self.file_location, "r", format='NETCDF3_CLASSIC')
		
		self.ionization_type = self.net_cdf_obj.test_ionization_mode

		self.experiment_type = self.net_cdf_obj.experiment_type

		self.list_scans = range(len(self.net_cdf_obj.variables.get("actual_scan_number")[:]))

		self.initial_scan_number = self.list_scans[0]

		self.final_scan_number = self.list_scans[-1]

		self.analyzer = analyzer

		self.instrument_label = instrument_label
		
		self.gcms = GCMSBase(self.file_location, analyzer, instrument_label)

		#other unused variables
		#a_d_sampling_rate = self.net_cdf_obj.variables.get("a_d_sampling_rate")[:]
		#a_d_coaddition_factor = self.net_cdf_obj.variables.get("a_d_coaddition_factor")[:]
		#inter_scan_time = self.net_cdf_obj.variables.get("inter_scan_time")[:]
		#flag_count = self.net_cdf_obj.variables.get("flag_count")[:]
		#instrument_name = ''.join(self.net_cdf_obj.variables.get("instrument_name")[:].tolist())
		#instrument_id = self.net_cdf_obj.variables.get("instrument_id")[:]
		#instrument_mfr = self.net_cdf_obj.variables.get("instrument_mfr")[:]
		#instrument_model = self.net_cdf_obj.variables.get("instrument_model")[:]
		#instrument_serial_no = self.net_cdf_obj.variables.get("instrument_serial_no")[:]
		#instrument_sw_version = self.net_cdf_obj.variables.get("instrument_sw_version")[:]
		#instrument_fw_version = self.net_cdf_obj.variables.get("instrument_fw_version")[:]
		#instrument_os_version = self.net_cdf_obj.variables.get("instrument_os_version")[:]
		#instrument_app_version = self.net_cdf_obj.variables.get("instrument_app_version")[:]
		#instrument_comments = self.net_cdf_obj.variables.get("instrument_comments")[:]
		#scan_duration = self.net_cdf_obj.variables.get("scan_duration")[:]
		#time_range_min = self.net_cdf_obj.variables.get("time_range_min")[:]
		#time_range_max = self.net_cdf_obj.variables.get("time_range_max")[:]
		#scan_index_list = self.net_cdf_obj.variables.get("scan_index")[:]
		#time_values = self.net_cdf_obj.variables.get("time_values")[:]
		
	@property

	def polarity(self):

		polarity = str(self.net_cdf_obj.test_ionization_polarity)

		if polarity == 'Positive Polarity': return +1
			
		else: return -1    

	def get_mass_spectrum(self, mz, abun, rp, d_params):
				
		data_dict = {Labels.mz: mz,
				Labels.abundance: abun,
				Labels.rp: rp,
				Labels.s2n: None,
		}
            
		mass_spec = MassSpecCentroidLowRes(data_dict, d_params)
		
		self.gcms.add_mass_spectrum(mass_spec)
		
	def run(self):
        
		'''populate the gcms obj'''

		d_parameters = default_parameters(self.file_location)
		
		self.import_mass_spectra(d_parameters)
	
	def import_mass_spectra(self, d_params):
		
		ms_datapoints_per_scans = self.net_cdf_obj.variables.get("point_count")[:]
		
		list_tic = self.net_cdf_obj.variables.get("total_intensity")[:]	

		list_rt = self.net_cdf_obj.variables.get("scan_acquisition_time")[:]/60

		mass_values = self.net_cdf_obj.variables.get("mass_values")[:]
		
		intensity_values = self.net_cdf_obj.variables.get("intensity_values")[:]
		
		resolution = self.net_cdf_obj.variables.get("resolution")[:]

		individual_rp = len(mass_values) == len(resolution)

		finish_location = -1
		
		for scan_index in self.list_scans:

			datapoints = ms_datapoints_per_scans[scan_index]

			finish_location += datapoints 

			start_location = finish_location - datapoints + 1

			#finish_location = ms_datapoints_per_scans[start_location:last_pos]
						
			d_params["rt"] = list_rt[scan_index]

			d_params["scan_number"] = scan_index

			d_params['label'] = Labels.gcms_centroid

			d_params["polarity"] = self.polarity

			d_params['analyzer'] = self.analyzer

			d_params['instrument_label'] = self.instrument_label

			mz = mass_values[start_location:finish_location]
			
			abun = intensity_values[start_location:finish_location]

			if individual_rp: 
				
				rp = resolution[start_location:finish_location]
			
			else:
				rp = [resolution[scan_index]] * datapoints

			#t1 = threading.Thread(target=self.get_mass_spectrum, args=(mz, abun, rp, d_params))
			#t1.start()
			
			self.get_mass_spectrum( mz, abun, rp, d_params)

		
		self.gcms.retention_time = list_rt
		self.gcms.tic = list_tic
		self.gcms.scans_number = self.list_scans	
	
	def get_gcms_obj(self):
        
		return self.gcms
		


__author__ = "Yuri E. Corilo"
__date__ = "Jun 04, 2019"

from scipy.stats import norm, cauchy
from numpy import linspace, sqrt, log, trapz, pi, log, poly1d, polyfit
from corems.encapsulation.constant import Atoms
from lmfit import models

class MSPeakCalculation(object):

    '''
    classdocs
    '''

    def _calc_kdm(self, dict_base):
        '''dict_base = {"C": 1, "H": 2}
        '''
        
        mass = 0
        for atom in dict_base.keys():
            mass += Atoms.atomic_masses.get(atom) * dict_base.get(atom)

        kendrick_mass = (int(mass) / mass) * self.mz_exp

        nominal_km = int(kendrick_mass)

        kmd = (nominal_km - kendrick_mass) 

        # kmd = (nominal_km - km) * 1
        #kdm = round(kmd,0)

        return kmd, kendrick_mass, nominal_km

    def calc_area(self, dx=1):
        
        if self.final_index > self.start_index:
        
            yy = self._ms_parent.abundance_profile[self.start_index:self.final_index]
            
            return trapz(yy, dx = dx)
        
        else:
            
            return None
            
    def voigt(self, oversample_multiplier=1, delta_rp = 0, mz_overlay=1):
        
        
        
        if self.resolving_power:

            # full width half maximum distance
            self.fwhm = (self.mz_exp / (self.resolving_power + delta_rp))#self.resolving_power)

            # stardart deviation
            sigma = self.fwhm / 3.6013

            # half width baseline distance
            
            #mz_domain = linspace(self.mz_exp - hw_base_distance,
            #                     self.mz_exp + hw_base_distance, datapoint)
            mz_domain = self.get_mz_domain(oversample_multiplier, mz_overlay)    
            
            # gaussian_pdf = lambda x0, x, s: (1/ math.sqrt(2*math.pi*math.pow(s,2))) * math.exp(-1 * math.pow(x-x0,2) / 2*math.pow(s,2) )
            
            #TODO derive amplitude
            amplitude = (sqrt(2*pi)*sigma) * self.abundance

            model = models.VoigtModel()

            params = model.make_params(center=self.mz_exp, amplitude=amplitude, sigma = sigma, gamma = sigma)

            calc_abundance = model.eval(params=params, x=mz_domain)

            return mz_domain, calc_abundance
        
        else:
            
            raise LookupError(
                'resolving power is not defined, try to use set_max_resolving_power()')

    def pseudovoigt(self, oversample_multiplier=1, delta_rp = 0, mz_overlay=1, fraction =0.5):
        
        if self.resolving_power:

            # full width half maximum distance
            self.fwhm = (self.mz_exp / (self.resolving_power + delta_rp))#self.resolving_power)

            # stardart deviation
            sigma = self.fwhm / 2

            # half width baseline distance
            
            #mz_domain = linspace(self.mz_exp - hw_base_distance,
            #                     self.mz_exp + hw_base_distance, datapoint)
            mz_domain = self.get_mz_domain(oversample_multiplier, mz_overlay)    
            
            # gaussian_pdf = lambda x0, x, s: (1/ math.sqrt(2*math.pi*math.pow(s,2))) * math.exp(-1 * math.pow(x-x0,2) / 2*math.pow(s,2) )
            model = models.PseudoVoigtModel()
            
            # TODO derive amplitude
            gamma = sigma
            
            amplitude = (sqrt(2*pi)*sigma) * self.abundance
            amplitude = (sqrt(pi/log(2)) * (pi*sigma*self.abundance)) /( (pi*(1-gamma)) + (sqrt(pi*log(2)) * gamma) )

            params = model.make_params(center=self.mz_exp, sigma = sigma)

            calc_abundance = model.eval(params=params, x=mz_domain)

            return mz_domain, calc_abundance
        
        else:
            
            raise LookupError(
                'resolving power is not defined, try to use set_max_resolving_power()')

    def get_mz_domain(self, oversample_multiplier, mz_overlay):
        
        start_index = self.start_index - mz_overlay  if not self.start_index == 0 else 0
        final_index = self.final_index + mz_overlay  if not self.final_index == len(self._ms_parent.mz_exp_profile) else self.final_index

        if oversample_multiplier == 1:

            mz_domain = self._ms_parent.mz_exp_profile[start_index: final_index]
            
        else:
            # we assume a linear correlation for m/z and datapoits 
            # which is only true if the m/z range in narrow (within 1 m/z unit)
            # this is not true for a wide m/z range
                         
            indexes = range(start_index, final_index+1)
            mz = self._ms_parent.mz_exp_profile[indexes]
            pol = poly1d(polyfit(indexes, mz, 1))
            oversampled_indexes = linspace(start_index, final_index, (final_index-start_index) * oversample_multiplier)    
            mz_domain = pol(oversampled_indexes)

        return mz_domain

    def lorentz(self, oversample_multiplier=1, delta_rp = 0, mz_overlay=1):

        if self.resolving_power:

            # full width half maximum distance
            self.fwhm = (self.mz_exp / (self.resolving_power + delta_rp))#self.resolving_power)

            # stardart deviation
            sigma = self.fwhm / 2

            # half width baseline distance
            hw_base_distance = (8 * sigma)

            #mz_domain = linspace(self.mz_exp - hw_base_distance,
            #                     self.mz_exp + hw_base_distance, datapoint)
            
            
            mz_domain = self.get_mz_domain(oversample_multiplier, mz_overlay)    
            # gaussian_pdf = lambda x0, x, s: (1/ math.sqrt(2*math.pi*math.pow(s,2))) * math.exp(-1 * math.pow(x-x0,2) / 2*math.pow(s,2) )
            model = models.LorentzianModel()
            
            amplitude = sigma* pi * self.abundance

            params = model.make_params(center=self.mz_exp, amplitude=amplitude, sigma = sigma)

            calc_abundance = model.eval(params=params, x=mz_domain)

            return mz_domain, calc_abundance
        
        else:
            
            raise LookupError(
                'resolving power is not defined, try to use set_max_resolving_power()')

    def gaussian(self, oversample_multiplier=1, delta_rp = 0, mz_overlay=1):
        

        # check if MSPeak contains the resolving power info
        if self.resolving_power:
            # full width half maximum distance
            self.fwhm = (self.mz_exp / (self.resolving_power + delta_rp))#self.resolving_power)

            # stardart deviation
            sigma = self.fwhm / (2 * sqrt(2 * log(2)))

            # half width baseline distance
            #hw_base_distance = (3.2 * s)

            #match_loz_factor = 3

            #n_d = hw_base_distance * match_loz_factor

            #mz_domain = linspace(
            #    self.mz_exp - n_d, self.mz_exp + n_d, datapoint)

            mz_domain = self.get_mz_domain(oversample_multiplier, mz_overlay)    
            
            # gaussian_pdf = lambda x0, x, s: (1/ math.sqrt(2*math.pi*math.pow(s,2))) * math.exp(-1 * math.pow(x-x0,2) / 2*math.pow(s,2) )
            
            #calc_abundance = norm.pdf(mz_domain, self.mz_exp, s)

            model = models.GaussianModel()
            
            amplitude = (sqrt(2*pi)*sigma) * self.abundance

            params = model.make_params(center=self.mz_exp, amplitude=amplitude, sigma = sigma)

            calc_abundance = model.eval(params=params, x=mz_domain)
            
            return mz_domain, calc_abundance 

        else:
            raise LookupError(
                'resolving power is not defined, try to use set_max_resolving_power()')
    
    @property
    def number_possible_assignments(self,):
        
        return len(self.molecular_formulas)

    def molecular_formula_lowest_error(self):
       
       return min(self.molecular_formulas, key=lambda m: abs(m.mz_error))

    def molecular_formula_highest_prob_score(self):
       
       return max(self.molecular_formulas, key=lambda m: abs(m.confidence_score))

    def molecular_formula_earth_filter(self, lowest_error=True):
        
        candidates = list(filter(lambda mf: mf.get("O") > 0 and mf.get("N") <=3 and mf.get("P") <= 2 and (3 * mf.get("P")) <= mf.get("O"), self.molecular_formulas))

        if lowest_error:
            return min(candidates, key=lambda m: abs(m.mz_error))
        else:
            return candidates

    def molecular_formula_water_filter(self, lowest_error=True):
       
        candidates = list(filter(lambda mf: mf.get("O") > 0 and mf.get("N") <=3 and mf.get("S") <=2 and  mf.get("P") <= 2, self.molecular_formulas))

        if lowest_error:
            return min(candidates, key=lambda m: abs(m.mz_error))
        else:
            return candidates
    
    def molecular_formula_air_filter(self, lowest_error=True):
       
        candidates = list(filter(lambda mf: mf.get("O") > 0 and mf.get("N") <=2 and mf.get("S") <=1 and  mf.get("P") == 0 and 3* (mf.get("S") + mf.get("N")) <= mf.get("O"), self.molecular_formulas))
        
        if lowest_error:
            return min(candidates, key=lambda m: abs(m.mz_error))
        else:
            return candidates

    def cia_score_S_P_error(self):
        #case EFormulaScore.HAcap:

        lowest_S_P_mf = min(self.molecular_formulas, key=lambda mf: mf.get('S') + mf.get('P'))
        lowest_S_P_count = lowest_S_P_mf.get("S") + lowest_S_P_mf.get("P")
        
        list_same_s_p = list(filter(lambda mf: mf.get('S') + mf.get('P') == lowest_S_P_count, self.molecular_formulas))

        #check if list is not empty
        if list_same_s_p:
        
            return min(list_same_s_p, key=lambda m: abs(m.mz_error))
        
        else:
        
            return lowest_S_P_mf
    
    def cia_score_N_S_P_error(self):
        #case EFormulaScore.HAcap:
        if self.molecular_formulas:

            lowest_N_S_P_mf = min(self.molecular_formulas, key=lambda mf: mf.get('N') + mf.get('S') + mf.get('P'))
            lowest_N_S_P_count = lowest_N_S_P_mf.get("N") + lowest_N_S_P_mf.get("S") + lowest_N_S_P_mf.get("P")

            list_same_N_S_P = list(filter(lambda mf: mf.get('N') + mf.get('S') + mf.get('P') == lowest_N_S_P_count, self.molecular_formulas))

            if list_same_N_S_P:

                SP_filtered_list =  list(filter(lambda mf: (mf.get("S") <= 3 ) and  (mf.get("P")  <= 1 ), list_same_N_S_P))
                
                if SP_filtered_list:
                    
                    return min(SP_filtered_list, key=lambda m: abs(m.mz_error)) 
                
                else:    
                    
                    return min(list_same_N_S_P, key=lambda m: abs(m.mz_error))            
            
            else:
                
                return lowest_N_S_P_mf 
        else:
            raise Exception("No molecular formula associated with the mass spectrum peak at m/z: %.6f" % self.mz_exp)



import numpy
import multiprocessing
from threading import Thread
import sys
import site
from pathlib import Path
from io import BytesIO

import clr
import pandas as pd
from s3path import S3Path
from tqdm import tqdm


from corems.encapsulation.constant import Labels
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecProfile, MassSpecCentroid
from corems.mass_spectra.factory.LC_Class import LCMSBase
from corems.encapsulation.factory.parameters import default_parameters


# do not change the order from the imports statements and reference below
sys.path.append(site.getsitepackages()[0] + "/ext_lib")
sys.path.append("ext_lib")

clr.AddReference("ThermoFisher.CommonCore.RawFileReader")
clr.AddReference("ThermoFisher.CommonCore.Data")
clr.AddReference("ThermoFisher.CommonCore.MassPrecisionEstimator")

from ThermoFisher.CommonCore.RawFileReader import RawFileReaderAdapter
from ThermoFisher.CommonCore.Data import ToleranceUnits, Extensions
from ThermoFisher.CommonCore.Data.Business import MassOptions
from ThermoFisher.CommonCore.Data.FilterEnums import MSOrderType 
from System.Collections.Generic import List

class ImportMassSpectraThermoMSFileReader():

    """     Read FULL mode spectra only from raw file data and store it return a LC-MS class
    *  Default behavior is to load all scans numbers

    *  set start_scan_number  and final_scan_number to change it before calling start(), or run()
    """

    def __init__(self, file_location):

        # Thread.__init__(self)
        if isinstance(file_location, str):
            file_path = Path(file_location)

        elif isinstance(file_location, S3Path):

            temp_dir = Path('tmp/')
            temp_dir.mkdir(exist_ok=True)

            file_path = temp_dir / file_location.name 
            with open(file_path, 'wb') as fh:
                fh.write(file_location.read_bytes())
        else:
            file_path = file_location
        self.iRawDataPlus = RawFileReaderAdapter.FileFactory(str(file_path))

        #   removing tmp file

        if isinstance(file_location, S3Path):
            file_path.unlink()

        self.res = self.iRawDataPlus.SelectInstrument(0, 1)

        self._initial_scan_number = self.iRawDataPlus.RunHeaderEx.FirstSpectrum

        self._final_scan_number = self.iRawDataPlus.RunHeaderEx.LastSpectrum

        self.file_location = file_location

    @property
    def initial_scan_number(self):
        return self._initial_scan_number

    @property
    def final_scan_number(self):
        return self._final_scan_number

    def get_filter_for_scan_num(self, scan_number):
        """
        Returns the closest matching run time that corresponds to scan_number for the current
        controller. This function is only supported for MS device controllers.
        e.g.  ['FTMS', '-', 'p', 'NSI', 'Full', 'ms', '[200.00-1000.00]']
        """
        scan_label = self.iRawDataPlus.GetScanEventStringForScanNumber(
            scan_number)

        return str(scan_label).split()

    def get_all_filters(self):
        """
        Get all scan filters.
        This function is only supported for MS device controllers.
        e.g.  ['FTMS', '-', 'p', 'NSI', 'Full', 'ms', '[200.00-1000.00]']
        """
        scanrange = range(self._initial_scan_number, self._final_scan_number + 1)
        scanfiltersdic = {}
        scanfilterslist = []
        for scan_number in scanrange:
            scan_label = self.iRawDataPlus.GetScanEventStringForScanNumber(scan_number)
            scanfiltersdic[scan_number] = scan_label
            scanfilterslist.append(scan_label)
        scanfilterset = list(set(scanfilterslist))
        return scanfiltersdic, scanfilterset

    def check_full_scan(self, scan_number):
        # scan_filter.ScanMode 0 = FULL
        scan_filter = self.iRawDataPlus.GetFilterForScanNumber(scan_number)

        return scan_filter.ScanMode == 0

    def get_polarity_mode(self, scan_number):

        polarity_symbol = self.get_filter_for_scan_num(scan_number)[1]

        if polarity_symbol == "+":

            return 1
            # return "POSITIVE_ION_MODE"

        elif polarity_symbol == "-":

            return -1

        else:

            raise Exception("Polarity Mode Unknown, please set it manually")

    def get_scan_header(self, scan):
        '''
        Get full dictionary of scan header meta data, i.e. AGC status, ion injection time, etc.
        '''
        header = self.iRawDataPlus.GetTrailerExtraInformation(scan)
        header_dic = {}
        for i in numpy.arange(header.Length):
            header_dic.update({header.Labels[i]: header.Values[i]})
        return header_dic

    def get_icr_transient_times(self, first_scan: int = None, last_scan: int = None):
        '''
        Return a list for transient time targets for all scans, or selected scans range
        Resolving Power and Transient time targets based on 7T FT-ICR MS system
        '''

        res_trans_time = {"50": 0.384,
                          "100000": 0.768,
                          "200000": 1.536,
                          "400000": 3.072,
                          "750000": 6.144,
                          "1000000": 12.288
                          }

        firstScanNumber = self._initial_scan_number if first_scan is None else first_scan

        lastScanNumber = self._final_scan_number if last_scan is None else last_scan

        transient_time_list = []

        for scan in range(firstScanNumber, lastScanNumber):

            scan_header = self.get_scan_header(scan)

            rp_target = scan_header['FT Resolution:']

            transient_time = res_trans_time.get(rp_target)

            transient_time_list.append(transient_time)

            # print(transient_time, rp_target)

        return transient_time_list

    def get_data(self, scan, d_parameter, scan_type):

        if scan_type == "Centroid":

            centroidStream = self.iRawDataPlus.GetCentroidStream(scan, False)

            noise = list(centroidStream.Noises)

            baselines = list(centroidStream.Baselines)

            rp = list(centroidStream.Resolutions)

            magnitude = list(centroidStream.Intensities)

            mz = list(centroidStream.Masses)

            # charge = scans_labels[5]
            array_noise_std = (numpy.array(noise) - numpy.array(baselines)) / 3
            l_signal_to_noise = numpy.array(magnitude) / array_noise_std

            d_parameter["baselise_noise"] = numpy.average(array_noise_std)

            d_parameter["baselise_noise_std"] = numpy.std(array_noise_std)

            data_dict = {
                Labels.mz: mz,
                Labels.abundance: magnitude,
                Labels.rp: rp,
                Labels.s2n: l_signal_to_noise,
            }

        else:

            scanStatistics = self.iRawDataPlus.GetScanStatsForScanNumber(scan)

            profileStream = self.iRawDataPlus.GetSegmentedScanFromScanNumber(
                scan, scanStatistics)

            magnitude = list(profileStream.Intensities)

            mz = list(profileStream.Positions)

            data_dict = {
                Labels.mz: mz,
                Labels.abundance: magnitude,
            }

        return data_dict

    def set_metadata(self, firstScanNumber=0, lastScanNumber=0, scans_list=False):

        d_params = default_parameters(self.file_location)

        # assumes scans is full scan or reduced profile scan

        d_params["label"] = Labels.thermo_profile

        if scans_list:
            d_params['scan_number'] = scans_list

            d_params["polarity"] = self.get_polarity_mode(scans_list[0])

        else:

            d_params['scan_number'] = "{}-{}".format(firstScanNumber, lastScanNumber)

            d_params["polarity"] = self.get_polarity_mode(firstScanNumber)

        d_params['analyzer'] = self.iRawDataPlus.GetInstrumentData().Model

        d_params['instrument_label'] = self.iRawDataPlus.GetInstrumentData().Name

        return d_params

    def get_average_mass_spectrum_by_scanlist(self, scans_list, auto_process: bool = True, ppm_tolerance: float = 5.0):

        d_params = self.set_metadata(scans_list=scans_list)

        # assumes scans is full scan or reduced profile scan

        scans = List[int]()
        for scan in scans_list:
            scans.Add(scan)

        # Create the mass options object that will be used when averaging the scans
        options = MassOptions()
        options.ToleranceUnits = ToleranceUnits.ppm
        options.Tolerance = ppm_tolerance

        # Get the scan filter for the first scan.  This scan filter will be used to located
        # scans within the given scan range of the same type

        averageScan = Extensions.AverageScans(self.iRawDataPlus, scans, options)

        len_data = averageScan.SegmentedScan.Positions.Length

        mz_list = list(averageScan.SegmentedScan.Positions)
        abund_list = list(averageScan.SegmentedScan.Intensities)

        data_dict = {
                    Labels.mz: mz_list,
                    Labels.abundance: abund_list,
                }

        mass_spec = MassSpecProfile(data_dict, d_params, auto_process=auto_process)

        return mass_spec

    def get_average_mass_spectrum_in_scan_range(self, first_scan: int = None, last_scan: int = None, auto_process: bool = True, ppm_tolerance: float = 5.0, ms_type=MSOrderType.Ms):

        firstScanNumber = self._initial_scan_number if first_scan is None else first_scan

        lastScanNumber = self._final_scan_number if last_scan is None else last_scan

        d_params = self.set_metadata(firstScanNumber=firstScanNumber, lastScanNumber=lastScanNumber)

        # Create the mass options object that will be used when averaging the scans
        options = MassOptions()

        options.ToleranceUnits = ToleranceUnits.ppm
        options.Tolerance = ppm_tolerance

        # Get the scan filter for the first scan.  This scan filter will be used to located
        # scans within the given scan range of the same type
        scanFilter = self.iRawDataPlus.GetFilterForScanNumber(firstScanNumber)

        # force it to only look for the MSType
        scanFilter.MSOrder = ms_type

        averageScan = Extensions.AverageScansInScanRange(self.iRawDataPlus, firstScanNumber, lastScanNumber, scanFilter, options)

        if averageScan:
            mz_list = list(averageScan.SegmentedScan.Positions)
            abund_list = list(averageScan.SegmentedScan.Intensities)        

            data_dict = {
                        Labels.mz: mz_list,
                        Labels.abundance: abund_list,
                    }

            mass_spec = MassSpecProfile(data_dict, d_params, auto_process=auto_process)

            return mass_spec
        else:
            raise Exception('no data found for the MSOrderType = {}'.format(ms_type))

    def get_summed_mass_spectrum(self, initial_scan_number, final_scan_number=None,
                                 auto_process=True, pd_method=True, pd_merge_n=100):

        d_params = default_parameters(self.file_location)

        # assumes scans is full scan or reduced profile scan

        d_params["label"] = Labels.thermo_profile

        if type(initial_scan_number) is list:
            d_params["polarity"] = self.get_polarity_mode(initial_scan_number[0])

            scanrange = initial_scan_number
        else:
            d_params["polarity"] = self.get_polarity_mode(initial_scan_number)

            if final_scan_number is None:
                final_scan_number = self._final_scan_number

            scanrange = range(initial_scan_number, final_scan_number + 1)

        if pd_method:

            def sort_sum_df(df):
                """
                Nested function to sort dataframe and sum rows with exact matching indexes (m/z)
                """
                df = df.sort_index()
                df = df.groupby(level=0).sum()
                return df

            # initialise empty Pandas series
            big_df = pd.Series(index=[], dtype='float64')

            for scan_number in tqdm(scanrange):
                scanStatistics = self.iRawDataPlus.GetScanStatsForScanNumber(scan_number)
                segmentedScan = self.iRawDataPlus.GetSegmentedScanFromScanNumber(scan_number, scanStatistics)

                tmp_df = pd.Series(index=list(segmentedScan.Positions),
                                    dtype='float64', data=list(segmentedScan.Intensities))
                big_df = big_df.append(tmp_df)

                # this allows you to merge/sum the values earlier, however it slows down a lot
                # limited benefit unless running into memory issues
                # for complex data it is necessary to stop the iterations getting too slow
                if scan_number % pd_merge_n == 0:
                    big_df = sort_sum_df(big_df)

            big_df = sort_sum_df(big_df)
            data_dict = {
                        Labels.mz: list(big_df.index.values),
                        Labels.abundance: list(big_df.values),
                    }
        else:
            all_mz = dict()

            for scan_number in tqdm(scanrange):

                scanStatistics = self.iRawDataPlus.GetScanStatsForScanNumber(scan_number)

                segmentedScan = self.iRawDataPlus.GetSegmentedScanFromScanNumber(scan_number, scanStatistics)

                len_data = segmentedScan.Positions.Length

                for i in range(len_data):

                    mz = segmentedScan.Positions[i]
                    abundance = segmentedScan.Intensities[i]

                    if mz in all_mz:
                        all_mz[mz] = all_mz[mz] + abundance
                    else:
                        all_mz[mz] = abundance

            mz_all = []
            abun_all = []

            for mz in sorted(all_mz):
                mz_all.append(mz)
                abun_all.append(all_mz[mz])

            data_dict = {
                        Labels.mz: mz_all,
                        Labels.abundance: abun_all,
                    }

        print('Summed. Now Processing.')

        mass_spec = MassSpecProfile(data_dict, d_params, auto_process=auto_process)

        return mass_spec

    def get_tic(self,plot=False):
        """
        Reads the TIC values for each scan from the Thermo headers
        Returns a pandas dataframe of Scans, TICs, and Times
        (Optionally) plots the TIC chromatogram.
        """
        first_scan = self._initial_scan_number
        final_scan = self._final_scan_number
        scanrange = range(first_scan,final_scan+1)

        ms_tic = pd.DataFrame(index=scanrange,columns=['Time','TIC'])
        for scan in scanrange:
            scanStatistics = self.iRawDataPlus.GetScanStatsForScanNumber(scan)
            ms_tic.loc[scan,'Time'] = scanStatistics.StartTime
            ms_tic.loc[scan,'TIC'] = scanStatistics.TIC

        
        if plot:
            import matplotlib.pyplot as plt #maybe better in top of file?
            fig,ax = plt.subplots(figsize=(6,3))
            ax.plot(ms_tic['Time'],ms_tic['TIC'],label='TIC')
            ax.set_xlabel('Time (min)')
            ax.set_ylabel('a.u.')
            plt.legend()
            # plt.show()
            return ms_tic,fig
        return ms_tic

    def get_best_scans_idx(self, stdevs=2, method='mean', plot=False):
        '''
        Method to determine the best scan indexes for selective co-addition
        Based on calculating the mean (default) of the TIC values
        and setting an upper limit above/below that within X standard deviations.
        Mean or median makes limited difference, it seems.
        Empirically, 1-2 stdevs enough to filter out the worst datapoints.
        Optionally, plot the TIC with horizontal lines for the standard dev cutoffs.
        '''
        tic = self.get_tic()

        if method == 'median':
            tic_median = tic['TIC'].median()
        elif method == 'mean':
            tic_median = tic['TIC'].mean()
        else:
            print("Method " + str(method) + " undefined")

        tic_std = tic['TIC'].std()

        upperlimit = tic_median - (stdevs * tic_std)
        lowerlimit = tic_median + (stdevs * tic_std)

        tic_filtered = tic[(tic['TIC'] > upperlimit) & (tic['TIC'] < lowerlimit)]
        scans = list(tic_filtered.index.values)

        if plot:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(figsize=(8, 4))
            ax.plot(tic['Time'], tic['TIC'])
            ax.axhline(y=upperlimit, c='r')
            ax.axhline(y=lowerlimit, c='r')
            return fig, scans
        else:
            return scans



from ctypes import c_long
from threading import Thread

from comtypes.automation import BSTR
from comtypes.client import CreateObject
from numpy import array


from corems.encapsulation.factory.parameters import default_parameters
from corems.encapsulation.constant import Labels
from corems.mass_spectra.factory.LC_Class import LCMSBase
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecProfile

__author__ = "Yuri E. Corilo"
__date__ = "July 10, 2019"

class ImportLCMSBrukerCompassXtract(Thread):
    '''class docs'''
    
    def __init__(self, file_location, auto_process=True):

        Thread.__init__(self)

        self.lcms = LCMSBase(file_location)

        """Set up the COM object interface"""
        self.Bruker_Library = CreateObject("EDAL.MSAnalysis")

        self.res = self.Bruker_Library.Open(file_location)

        self.check_load_sucess()

        self._initial_scan_number = 1

        self._final_scan_number = self.get_scans_numbers()

        self.file_location = file_location

        self.auto_process = auto_process

    @property
    def initial_scan_number(self):
        return self._initial_scan_number

    @property
    def final_scan_number(self):
        return self._final_scan_number

    def check_scan(self, scan):

        scan_numbers = self.get_scans_numbers()
        return scan <= scan_numbers

    @initial_scan_number.setter
    def initial_scan_number(self, initial_scan_number):
        if self.check_scan(initial_scan_number):
            self._initial_scan_number = initial_scan_number
        else:
            raise Exception(
                "startscan and finalscan should be less than %s"
                % self.get_scans_numbers()
            )

    @final_scan_number.setter
    def final_scan_number(self, final_scan_number):

        if self.check_scan(final_scan_number):
            self._final_scan_number = final_scan_number
        else:
            raise Exception(
                "startscan and finalscan should be less than %s"
                % self.get_scans_numbers()
            )

    def get_scans_numbers(self):

        scan_numbers = self.Bruker_Library.MSSpectrumCollection.Count

        return scan_numbers

    def get_polarity_mode(self, spectrum):

        polarity_symbol = spectrum.Polarity

        if polarity_symbol == 0:

            return 1
            # return "POSITIVE_ION_MODE"

        elif polarity_symbol == 1:

            return -1
            # return "NEGATIVE_ION_MODE"

        else:

            raise IOError("Could not read mass spectrum polarity mode")

    def check_load_sucess(self):
        """ 0 if successful; otherwise, see Error Codes """

        if self.res == 0:

            self.break_it = False

        else:

            raise ImportError(str(self.res))

    def get_bruker_tics(self):

        strAnalysisData = BSTR("SumIntensity")

        if self.Bruker_Library.HasAnalysisData(strAnalysisData):

            tics_array = self.Bruker_Library.GetAnalysisData(strAnalysisData)

            tics_array = array(tics_array)

        return tics_array

    def get_bruker_retention_time(self):

        strAnalysisData = BSTR("RetentionTime")

        if self.Bruker_Library.HasAnalysisData(strAnalysisData):

            tics_array = self.Bruker_Library.GetAnalysisData(strAnalysisData)

            tics_array = array(tics_array)
        else:
            tics_array = [0]

        return tics_array

    @staticmethod
    def get_data(spectra, scan):
        """init_variable_from_get_spectrums
        # massList set up later
        #retention_time = spectrum.RetentionTime
        """

        spectrum = spectra[scan]

        is_profile = c_long(1)

        masslist = spectrum.GetMassIntensityValues(is_profile)

        # index_to_cut = self.find_index_of_mass(1200, masslist[0])

        data_dict = {
            Labels.mz: array(masslist[0]),
            Labels.abundance: array(masslist[1]),
            Labels.rp: None,
            Labels.s2n: None,
        }

        return data_dict

    def run(self):
        '''creates the lcms obj'''
        d_parameters = default_parameters(self.file_location)
        self._import_mass_spectra(d_parameters)

    def _import_mass_spectra(self, d_params):

        spectra = self.Bruker_Library.MSSpectrumCollection

        list_rt = self.get_bruker_retention_time()

        list_Tics = self.get_bruker_tics()

        list_scans = list()

        for scan_number in range(self.initial_scan_number, self.final_scan_number + 1):

            if spectra[scan_number].MSMSStage == 1:
                # this label needs to go inside a encapsulation class for consistence
                d_params["label"] = Labels.bruker_profile

                d_params["polarity"] = self.get_polarity_mode(spectra[scan_number])

                d_params["rt"] = list_rt[scan_number - 1]

                d_params["scan_number"] = scan_number

                list_scans.append(scan_number)

                data_dict = self.get_data(spectra, scan_number)

                mass_spec = MassSpecProfile(data_dict, d_params, auto_process=self.auto_process)
                
                mass_spec.process_mass_spec()
                
                self.lcms.add_mass_spectrum(mass_spec)

        self.lcms.retention_time = list_rt
        self.lcms.tic = list_Tics
        self.lcms.scans_number = list_scans 
        # return each_mass_spectrum

    def get_lcms_obj(self):
        """get_lc_ms_class method should only be used when using this class as a Thread, 
        otherwise use the run() method to return the lcms class"""

        if self.lcms.get(self._initial_scan_number):
            return self.lcms
        else:
            raise Exception("returning a empty lcms class")



from copy import deepcopy
from collections import OrderedDict
from corems.molecular_formula.calc.MolecularFormulaCalc import MolecularFormulaCalc
from corems.encapsulation.constant import Atoms, Labels
from numpy import nan
import re

__author__ = "Yuri E. Corilo"
__date__ = "Jun 24, 2019"

class MolecularFormula(MolecularFormulaCalc):
    '''
    classdocs
    '''
    def __init__(self, molecular_formula, ion_charge, ion_type=None, adduct_atom=None, mspeak_parent=None):
        
        #clear dictionary of atoms with 0 value
        
        if   type(molecular_formula) is dict:
                self._from_dict(molecular_formula, ion_type, adduct_atom)   
        
        elif type(molecular_formula) is list:
                self._from_list(molecular_formula, ion_type, adduct_atom)   
        
        elif type(molecular_formula) is str:
                self._from_str(molecular_formula, ion_type, adduct_atom)   

        
        self._ion_charge = ion_charge

        self._confidence_score = None        
        self._isotopologue_similarity = None
        self._mz_error_score = None
        self._mass_error_average_score = None

        self.is_isotopologue = False

        # parent mass spectrum peak obj instance
        self._mspeak_parent = mspeak_parent

        self.expected_isotopologues = []
        self.mspeak_mf_isotopologues_indexes = []
        
        if self._mspeak_parent:
            kendrick_dict_base = self._mspeak_parent._ms_parent.kendrick_base
        else:
            kendrick_dict_base = {'C':1, 'H':2}
        self._kdm, self._kendrick_mass, self._nominal_km = self._calc_kdm(
            kendrick_dict_base)  
        
    def __repr__(self):

        return "MolecularFormula({0},{1},ion type = {2}".format(self._d_molecular_formula, self.ion_charge, self.ion_type)
    
    def __str__(self):

        return "MolecularFormula {0}, ion_charge:{1}, ion type:{2}, m/z:{3} ".format(self.string, self.ion_charge, self.ion_type, self.mz_calc)
    
    def __len__(self):
        
        #crash if keys are not ordered
            return len(self._d_molecular_formula.keys())
        
    def __getitem__(self, atom):
        
            #atom = list(self._d_molecular_formula.keys())[position]
            if atom in self._d_molecular_formula.keys():
                return self._d_molecular_formula[atom]
            else:
                return 0
    
    def get(self, atom):
        
            #atom = list(self._d_molecular_formula.keys())[position]
            if atom in self._d_molecular_formula.keys():
                return self._d_molecular_formula[atom]
            else:
                return 0
                
    def _from_dict(self, molecular_formula, ion_type, adduct_atom):
        
        self._d_molecular_formula = {key:val for key, val in molecular_formula.items() if val != 0}
        if ion_type:
            self._d_molecular_formula[Labels.ion_type] = ion_type
        if adduct_atom:
            if adduct_atom in self._d_molecular_formula:
                self._d_molecular_formula[adduct_atom] += 1 
            else: self._d_molecular_formula[adduct_atom] = 1 

        
    @property
    def isotopologue_count_percentile(self, ):
        if not len(self.expected_isotopologues) == 0:
            return (len(self.mspeak_mf_isotopologues_indexes)/len(self.expected_isotopologues))*100
        else: 
            return 100

    def _from_list(self, molecular_formula_list, ion_type, adduct_atom):
        # list has to be in the format 
        #['C', 10, 'H', 21, '13C', 1, 'Cl', 1, etc]  
        self._d_molecular_formula = {}
        for each in range(0, len(molecular_formula_list),2):
            
            atoms_label =  molecular_formula_list[each]
            atoms_count = int(molecular_formula_list[each+1])
            if atoms_count > 0:
                self._d_molecular_formula[atoms_label] = int(atoms_count)
        
        self._d_molecular_formula[Labels.ion_type] = ion_type
        if adduct_atom:
            if adduct_atom in self._d_molecular_formula:
                self._d_molecular_formula[adduct_atom] += 1 
            else: self._d_molecular_formula[adduct_atom] = 1 

    def _from_str(self, molecular_formula_str,  ion_type, adduct_atom):
        # string has to be in the format 
        #'C10 H21 13C1 Cl1 37Cl1 etc'
        molecular_formula_list = molecular_formula_str.split(' ')
        final_formula = []
        for i in molecular_formula_list:
            atoms_count = self.split(Atoms.atoms_order, i)
            final_formula.extend(atoms_count)
        print(final_formula)
        self._from_list(final_formula, ion_type, adduct_atom)

    def split(self, delimiters, string, maxsplit=0): #pragma: no cover
        
        ''' does not work when formula has atoms with same caracaters:
            i.e - C10H21NNa
        '''
        regexPattern = '|'.join(map(re.escape, delimiters)) #pragma: no cover
        isotopes = re.findall(regexPattern, string) #pragma: no cover
        counts = re.split(regexPattern, string, maxsplit)  #pragma: no cover
       
        return [isotopes[0], int(counts[1])]

    @property
    def O_C(self): 
            
            if 'O' in self._d_molecular_formula.keys():
                return self._d_molecular_formula.get("O")/self._d_molecular_formula.get("C")
            else:
                return 0    
    
    @property
    def H_C(self): return self._d_molecular_formula.get("H")/self._d_molecular_formula.get("C")
    
    @property
    def dbe(self): return self._calc_dbe()
    
    @property
    def mz_nominal_calc(self): return int(self._calc_mz())

    @property    
    def mz_error(self): return self._calc_assignment_mass_error()

    @property
    def mz_calc(self): return self._calc_mz()

    @property
    def ion_type(self): 
        
        ion_type = self._d_molecular_formula.get(Labels.ion_type)
        if ion_type == Labels.protonated_de_ion:
            if self.ion_charge > 0: 
                return Labels.protonated
            else: 
                return Labels.de_protonated    
        else:
            return ion_type

    @ion_type.setter
    def ion_type(self, ion_type):
        if  ion_type in [Labels.protonated_de_ion, Labels.adduct_ion, Labels.radical_ion]:
            self._d_molecular_formula[Labels.ion_type] = ion_type
        else:
            raise TypeError("Ion type can only be: 'DE_OR_PROTONATED', 'RADICAL' or  'ADDUCT', not %s"%ion_type)   

    @property
    def ion_charge(self): return self._ion_charge
    
    @property
    def atoms(self): return [key for key in self._d_molecular_formula.keys() if key != Labels.ion_type]
    
    @property
    def confidence_score(self): 
        
        if not self._confidence_score:
            
            self._confidence_score = self._calc_confidence_score() 
        
        return self._confidence_score

    @property
    def isotopologue_similarity(self): 
        
        if not self._isotopologue_similarity:
           
           self._isotopologue_similarity = self._calc_isotopologue_confidence()  
       
        return self._isotopologue_similarity  
    
    @property
    def average_mz_error_score(self): 
        
        ''' includes the isotopologues'''
        
        if not self._mass_error_average_score:
           
           self._mass_error_average_score = self._calc_average_mz_score()  
        
        return self._mass_error_average_score

    @property
    def mz_error_score(self): 
        if not self._mz_error_score:
           
           self._mz_error_score = self._calc_mz_confidence()  
        
        return self._mz_error_score
    
    @property
    def kmd(self): return self._kdm

    @property
    def kendrick_mass(self): return self._kendrick_mass

    @property
    def knm(self): return self._nominal_km

    def change_kendrick_base(self, kendrick_dict_base):
        '''kendrick_dict_base = {"C": 1, "H": 2}'''
        self._kdm, self._kendrick_mass, self._nominal_km = self._calc_kdm(kendrick_dict_base)
                
    def isotopologues(self, min_abundance, current_mono_abundance, dynamic_range): 
        
        # this calculation ignores the hydrogen
        for mf in self._cal_isotopologues(self._d_molecular_formula, min_abundance, current_mono_abundance, dynamic_range ):
             
            yield MolecularFormulaIsotopologue(*mf, current_mono_abundance, self.ion_charge)
    
    def atoms_qnt(self,atom): 
        if atom in self._d_molecular_formula:
            return self._d_molecular_formula.get(atom)
        else:
            raise Warning('Could not find %s in this Molecular Formula object'%str(atom))
    
    def atoms_symbol(self, atom): 
        '''return the atom symbol without the mass number'''
        return ''.join([i for i in atom if not i.isdigit()])

    @property       
    def string(self):
        
        if self._d_molecular_formula:
            formula_srt = ''
            for atom in Atoms.atoms_order:
                if atom in self.to_dict().keys():
                    formula_srt += atom + str(int(self.to_dict().get(atom))) + ' '
            return formula_srt.strip()
        
        else:
            raise Exception("Molecular formula identification not performed yet")    
    
    @property
    def string_formated(self):
        
        SUB = str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")
        SUP = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")
        
        if self._d_molecular_formula:
            formula_srt = ''
            for atom in Atoms.atoms_order:
                if atom in self.to_dict().keys():
                    formula_srt += atom.translate(SUP) + str(int(self.to_dict().get(atom))).translate(SUB)
            return formula_srt
        
        else:
            raise Exception("Molecular formula identification not performed yet")    

    def to_dict(self):
        return self._d_molecular_formula

    def to_list(self):
        '''TODO ensure self._d_molecular_formula is a orderedDict'''
        
        if self._d_molecular_formula:
            formula_list = []    
            
            for atom, atom_number in self._d_molecular_formula.items():
    
                if atom != Labels.ion_type:
                    
                    formula_list.append(atom)
                    formula_list.append(atom_number)
    
            return formula_list
        else:
            raise Exception("Molecular formula identification not performed yet")
    
    @property
    def class_label(self):
        
        if self._d_molecular_formula:
            
            formulalist = self.to_list()
            classstring = '' 
            
            for each in range(0, len(formulalist),2):
                
                if formulalist[each] != 'C' and formulalist[each] != 'H' and formulalist[each] != 'HC':
                     
                    classstring = classstring + str(formulalist[each]) + str(formulalist[each+1]) + ' '    
            
            if classstring == '': classstring = 'HC'
                
            classstring = classstring.strip()
            
            if self._d_molecular_formula.get(Labels.ion_type) == Labels.radical_ion:    
                
                return classstring + ' -R'
            
            #elif self._d_molecular_formula.get(Labels.ion_type) == Labels.adduct_ion:    
                
            #    return classstring + ' -A'

            else: return classstring
            
            #'dict, tuple or string'
        
        else:
            
            raise Exception("Molecular formula identification not performed yet")        
    
    @property
    def class_dict(self):
        
        if self._d_molecular_formula:
            
            class_dict = {}
            
            for atom, qnt in self._d_molecular_formula.items():
    
                if atom != Labels.ion_type and atom !='C' and atom !='H':
                    
                    class_dict[atom] = qnt
                    
            return class_dict
        
        raise Exception("Molecular formula identification not performed yet")           
    

class MolecularFormulaIsotopologue(MolecularFormula):
        
    '''
    classdocs
    '''
    def __init__(self, _d_molecular_formula, prob_ratio, mono_abundance, ion_charge, mspeak_parent=None):
        
        super().__init__(_d_molecular_formula,  ion_charge)
        #prob_ratio is relative to the monoisotopic peak p_isotopologue/p_mono_isotopic
        
        self.prob_ratio = prob_ratio
        
        self.abundance_calc = mono_abundance * prob_ratio

        self.is_isotopologue = True
        
        self.mspeak_index_mono_isotopic = None

        self.mono_isotopic_formula_index = None
        # parent mass spectrum peak obj instance
        self._mspeak_parent = mspeak_parent

    
    @property
    def area_error(self):
        return self._calc_area_error()

    @property
    def abundance_error(self):
        return self._calc_abundance_error()

        


__author__ = "Yuri E. Corilo"
__date__ = "Jun 27, 2019"

from numpy import power, multiply, sqrt, multiply, array, mean
from corems.mass_spectrum.calc.NoiseCalc import NoiseThresholdCalc
from corems.mass_spectrum.calc.PeakPicking import PeakPicking

class MassSpecCalc(PeakPicking, NoiseThresholdCalc ):
    '''
    Class including numerical calculations related to mass spectrum class
    Inherited PeakPicking and NoiseThresholdCalc ensuring its methods are 
    available to the instantiated mass spectrum class object
    '''
    def percentile_assigned(self, report_error=False):
        
        assign_abun = 0
        not_assign_abun = 0
        i = 0
        j = 0
        if report_error:
            error = []
        for mspeak in self.sort_by_abundance():
            
            if mspeak.is_assigned:
                i += 1
                assign_abun += mspeak.abundance
                if report_error:
                    error.append(mspeak.best_molecular_formula_candidate.mz_error)

            else:
                j += 1
                not_assign_abun += mspeak.abundance

        total_percent = (i / (i + j)) * 100
        total_relative_abundance = (assign_abun / (not_assign_abun + assign_abun)) * 100
        if report_error:
            rms_error = sqrt(mean(array(error)**2))
            print('%i assigned peaks and %i unassigned peaks, total  = %.2f %%, relative abundance = %.2f %%, RMS error (best candidate) (ppm) = %.3f' % (i, j, total_percent, total_relative_abundance, rms_error))
            return i, j, total_percent, total_relative_abundance, rms_error

        else:
            print('%i assigned peaks and %i unassigned peaks , total  = %.2f %%, relative abundance = %.2f %%' % (i, j, total_percent, total_relative_abundance,))
            return i, j, total_percent, total_relative_abundance

    def resolving_power_calc(self, B, T):
        '''
        low pressure limits, 
        T: float 
            transient time
        B: float
            Magnetic Filed Strength (Tesla)    
        
        reference
        Marshall et al. (Mass Spectrom Rev. 1998 Jan-Feb;17(1):1-35.)
        DOI: 10.1002/(SICI)1098-2787(1998)17:1<1::AID-MAS1>3.0.CO;2-K
        
        '''
       
        self.check_mspeaks()
        return array([mspeak.resolving_power_calc(B, T) for mspeak in self.mspeaks])
        
    def _f_to_mz(self):
        ''' Ledford equation for converting frequency(Hz) to m/z, 
        
        Attributes
        ----------
        All Attributes are derivative from the MassSpecBase Class
        by calling self
        
        Returns 
        ----------
            numpy.array(float)
            m/z domain after conversion from frequency
        '''
        Aterm, Bterm, Cterm = self.Aterm, self.Bterm, self.Cterm
        # Check if the Bterm of Ledford equation scales with the ICR trap voltage or not then Bterm = Bterm*trap_voltage
        
        if Cterm == 0:
            
            if Bterm == 0:
                #uncalibrated data
                mz_domain = Aterm / self.freq_exp_profile 
                
            else:
                
                mz_domain = (Aterm / (self.freq_exp_profile)) + (Bterm / power((self.freq_exp_profile), 2))

        # @will I need you insight here, not sure what is the inverted ledford equation that Bruker refers to
        else:

            mz_domain = (Aterm / self.freq_exp_profile) + (Bterm / power(self.freq_exp_profile, 2)) + Cterm

        return mz_domain

    def _f_to_mz_bruker(self):
        ''' 
        Burker equations for converting frequency (Hz) to m/z, 
        nOmega acquisition is not yet implemented here
        
        Attributes
        ----------
        All Attributes are derivative from the MassSpecBase Class
        
        Returns 
        ----------
            numpy.array(float)
            m/z domain after conversion from frequency
        '''
        Aterm, Bterm, Cterm = self.Aterm, self.Bterm, self.Cterm
        # Check if the Bterm of Ledford equation scales with the ICR trap voltage or not then Bterm = Bterm*trap_voltage
        if Cterm == 0:
            
            if Bterm == 0:
                #uncalibrated data
                return Aterm / self.freq_exp_profile 
            
            else:
                #calc2
                return Aterm / (self.freq_exp_profile + Bterm)

        # @will I need you insight here, not sure what is the inverted ledford equation that Bruker refers to
        else:
            diff = Aterm * Aterm
            
            #this sign(diff + 4) changes on older aquistion software
            diff = diff + 4 * Cterm * (self.freq_exp_profile - Bterm)
            diff = sqrt(diff)
            diff = -Aterm+diff
            #calc3
            return (2*Cterm)/diff
            return diff/2* (self.freq_exp_profile - Bterm)

    def number_average_molecular_weight(self, profile=False):
        ''' 
        Average molecular weight calculation 
        
        Attributes
        ----------
        All Attributes are derivative from the MassSpecBase Class
        
        Returns 
        ----------
            (float)
        '''
        # mode is profile or centroid data
        if profile:
            a = multiply(self.mz_exp_profile, self.abundance_profile)
            b = self.abundance_profile
            return a.sum()/b.sum()

        else:

            return sum(self.mz_exp*self.abundance)/sum(self.abundance)
    
    def weight_average_molecular_weight(self, profile=False):
        ''' 
        Weighted Average molecular weight calculation 
        
        Attributes
        ----------
        All Attributes are derivative from the MassSpecBase Class
        
        Returns 
        ----------
            (float)
        '''
        
        # implement from MassSpectralPeaks objs

        if profile:
            a = multiply(power(self.mz_exp_profile, 2), self.abundance_profile)
            b = self.mz_exp_profile*self.abundance_profile
            return a.sum() / b.sum()

        else:
            return sum(power(self.mz_exp, 2)*self.abundance)/sum(self.mz_exp*self.abundance)



__author__ = "Yuri E. Corilo"
__date__ = "Oct 29, 2019"


from threading import Thread

import h5py

from corems.encapsulation.constant import Labels
from corems.mass_spectrum.input.coremsHDF5 import ReadCoreMSHDF_MassSpectrum
from corems.mass_spectra.factory.LC_Class import LCMSBase

class ReadCoreMSHDF_MassSpectra(ReadCoreMSHDF_MassSpectrum, Thread):
    
    def __init__(self, file_location):
        
        Thread.__init__(self)
        
        ReadCoreMSHDF_MassSpectrum.__init__(self, file_location)
        
        self.lcms = LCMSBase(self.file_location)

        self.list_scans =  sorted([int(i) for i in list(self.h5pydata.keys())])

    def import_mass_spectra(self):
        
        list_rt, list_tic = list(), list()
        
        for scan_number in self.list_scans:
            
            mass_spec = self.get_mass_spectrum(scan_number)

            list_rt.append(mass_spec.rt)

            list_tic.append(mass_spec.tic)
            
            self.lcms.add_mass_spectrum(mass_spec)

        self.lcms.retention_time = list_rt
        self.lcms.tic = list_tic
        self.lcms.scans_number = self.list_scans
    
    def run(self):
        '''creates the lcms obj'''

        self.import_mass_spectra()
            
    def get_lcms_obj(self):
        
        if self.lcms:
            return self.lcms
        else:
            raise Exception("returning a empty lcms class")
    


__author__ = "Yuri E. Corilo"
__date__ = "Jan 31, 2020"

from collections.abc import Mapping

from matplotlib import pyplot as plt
from numpy import linspace

from corems.encapsulation.constant import Labels
from corems.encapsulation.constant import Atoms

flatten_list = lambda l: [item for sublist in l for item in sublist]

class HeteroatomsClassification(Mapping):
    
    '''Group mass spectrum data by heteroatom classes (Nn, Oo, Ss, NnOo, NnSs, etc..)
        
       class obj behaves as a dictionary of classes and return a list of ms_peak obj

    '''
    
    def __init__(self, mass_spectrum, choose_molecular_formula=True):

        def sort_atoms_method( atom):
            
            return [Atoms.atoms_order.index(atom)]

        self._ms_grouped_class = dict()
        
        self.choose_mf = choose_molecular_formula
        
        #mapping for ms peaks without any molecular formula associated
        self._ms_grouped_class[Labels.unassigned] = list()

        self.total_peaks = 0

        self.sum_abundance = 0

        self.min_max_mz = (mass_spectrum.min_mz_exp, mass_spectrum.max_mz_exp)

        self.min_max_abundance = (mass_spectrum.min_abundance, mass_spectrum.max_abundance)

        self.min_ppm_error = mass_spectrum.molecular_search_settings.min_ppm_error 

        self.max_ppm_error = mass_spectrum.molecular_search_settings.max_ppm_error

        check_assign = False

        all_used_atoms = set()

        for ms_peak in mass_spectrum:
            
            self.total_peaks += 1

            self.sum_abundance += ms_peak.abundance

            if not ms_peak.is_assigned:

                self._ms_grouped_class.get(Labels.unassigned).append(ms_peak)
                
            else:    
                
                check_assign = True    

                if choose_molecular_formula:
                    
                    mf = ms_peak.best_molecular_formula_candidate
                    
                    classes =  [mf.class_label]
                    
                    for atom in mf.atoms:
                        
                        all_used_atoms.add(atom)

                else: 

                    classes = []
                    
                    for mf in ms_peak:
                        
                        classes.append(mf.class_label)
                        
                        for atom in mf.atoms:
                             
                             all_used_atoms.add(atom)

                for classe in classes:
                    
                    if classe in self._ms_grouped_class.keys():

                        self._ms_grouped_class.get(classe).append(ms_peak)
                    
                    else:     

                        self._ms_grouped_class[classe] = [ms_peak]

        self.all_identified_atoms = sorted(all_used_atoms, key=sort_atoms_method)

        if not check_assign:

            raise Exception("No molecular formula associated with any mspeak objects")
    

    def __len__(self):
        
        return len(self._ms_grouped_class)
        
    def __getitem__(self, classe):
        
        return self._ms_grouped_class.get(classe)

    def __iter__(self):

         return iter(self._ms_grouped_class) 

    def get_classes(self, threshold_perc=1, isotopologue=True):
        
        classes = list()
        for classe in self.keys():
            if classe != Labels.unassigned:
                if self.abundance_count_percentile(classe) > threshold_perc:
                    
                    if classe != Labels.unassigned:
                        # access first molecular formula inside the first ms peak and check isotopologue
                        if not isotopologue and self.get(classe)[0][0].is_isotopologue: continue
                    
                    classes.append(classe)
        #TODO sort classes chemically here too
        return classes
        
    def molecular_formula_string(self, classe,):

        if self.choose_mf:
            return [mspeak.best_molecular_formula_candidate for mspeak in self[classe]]
        else:
            return [mf for mspeak in self[classe] for mf in mspeak if mf.class_label == classe]

    def molecular_formula(self, classe,):

        if self.choose_mf:
            return [mspeak.best_molecular_formula_candidate for mspeak in self[classe]]
        else:
            return [mf for mspeak in self[classe] for mf in mspeak if mf.class_label == classe]
            
    def carbon_number(self, classe):

        if self.choose_mf:
            return [mspeak.best_molecular_formula_candidate.get("C") for mspeak in self[classe]]
        else:
            return [mf.get('C') for mspeak in self[classe] for mf in mspeak if mf.class_label == classe]

    def atom_count(self, atom, classe):
        if self.choose_mf:
            return [mspeak.best_molecular_formula_candidate.get(atom) for mspeak in self[classe]]
        else:    
            return [mf.get(atom) for mspeak in self[classe] for mf in mspeak if mf.class_label == classe]

    def dbe(self, classe):
        if self.choose_mf:
            return [mspeak.best_molecular_formula_candidate.dbe for mspeak in self[classe]]
        else:    
            return [mf.dbe for mspeak in self[classe] for mf in mspeak if mf.class_label == classe]
    
    def atoms_ratio(self, classe, numerator, denominator):

        return [mf.get(numerator)/mf.get(denominator) for mf in self.molecular_formula(classe)]
       
    def mz_exp(self, classe):
        
        if self.choose_mf or classe == Labels.unassigned:
            
            return [mspeak.mz_exp for mspeak in self[classe]]
        
        else:
            
            return [mspeak.mz_exp for mspeak in self[classe] for mf in mspeak if mf.class_label == classe]
    
    def abundance(self, classe):

        if self.choose_mf or classe == Labels.unassigned:
            
            return [mspeak.abundance for mspeak in self[classe]]
        
        else:
            
            return [mspeak.abundance for mspeak in self[classe] for mf in mspeak if mf.class_label == classe]

    def mz_error(self, classe):

        if classe != Labels.unassigned:
            
            if self.choose_mf:
                
                return [mspeak.best_molecular_formula_candidate.mz_error for mspeak in self[classe]]
            
            else:
                
                return [mf.mz_error for mspeak in self[classe] for mf in mspeak if mf.class_label == classe]
    
    def mz_calc(self, classe):
        
        if self.choose_mf:
            
            return [mspeak.best_molecular_formula_candidate.mz_calc for mspeak in self[classe]]
        
        else:
            
            return [mf.mz_calc for mspeak in self[classe] for mf in mspeak if mf.class_label == classe] 

    def peaks_count_percentile(self, classe):

        return (len(self[classe])/self.total_peaks)*100

    def abundance_count_percentile(self, classe):

        return (sum([mspeak.abundance for mspeak in self[classe]])/self.sum_abundance)*100

    def mz_exp_assigned(self):

        classes = self.keys()

        return [mspeak.mz_exp for classe in classes for mspeak in self[classe] if classe != Labels.unassigned]
    
    def abundance_assigned(self):

        classes = self.keys()
            
        return [mspeak.abundance for classe in classes for mspeak in self[classe] if classe != Labels.unassigned]

    def mz_exp_all(self):
        
        classes = self.keys()
        
        return flatten_list([self.mz_exp(classe) for classe in classes if classe != Labels.unassigned])
    
    def mz_error_all(self):
        
        classes = self.keys()
        
        return flatten_list([self.mz_error(classe) for classe in classes if classe != Labels.unassigned])
        
    def carbon_number_all(self):

        classes = self.keys()
            
        return flatten_list([self.carbon_number(classe) for classe in classes if classe != Labels.unassigned])

    def dbe_all(self):

        classes = self.keys()
            
        return flatten_list([self.dbe(classe) for classe in classes if classe != Labels.unassigned])

    def atoms_ratio_all(self, numerator, denominator):

        classes = self.keys()
            
        return flatten_list([self.atoms_ratio(classe, numerator, denominator) for classe in classes if classe != Labels.unassigned])

    def to_dataframe(self, incluse_isotopologue=False, abundance_perc_threshold=5, include_unassigned=False):
        
        from pandas import DataFrame
        
        columns_labels = ['mz', 'calibrated_mz', 'calculated_m_z', 'abundance',
                                'resolving_power', 'sn', 'ion_charge', 'mass_error',
                                'DBE', 'class', 'HC', 'OC', 'ion_type','is_isotopologue',
                                'class_abundance', 'class_count']

        dict_data_list = []

        for classe, list_mspeaks in self.items():

            percent_abundance = self.abundance_count_percentile(classe)
            
            #ignores low abundant classes
            if abundance_perc_threshold < abundance_perc_threshold: continue
                
            peaks_count_percentile = self.peaks_count_percentile(classe)

            for ms_peak in list_mspeaks:
                 
                if ms_peak.is_assigned:
                    
                    for m_formula in ms_peak:
                        
                        #ignores isotopologues
                        if not incluse_isotopologue and m_formula.is_isotopologue: continue
                        
                        formula_dict = m_formula.to_dict()

                        dict_result = {'mz':  ms_peak._mz_exp,
                                'calibrated_mz': ms_peak.mz_exp,
                                'calculated_mz': m_formula.mz_calc,
                                'abundance': ms_peak.abundance,
                                'resolving_power': ms_peak.resolving_power,
                                'sn':  ms_peak.signal_to_noise,
                                'ion_charge': ms_peak.ion_charge,
                                'mass_error': m_formula.mz_error,
                                'DBE':  m_formula.dbe,
                                'class': classe,
                                'HC':  m_formula.H_C,
                                'OC':  m_formula.O_C,
                                'ion_type': str(m_formula.ion_type.lower().encode('utf-8')),
                                'is_isotopologue': int(m_formula.is_isotopologue),
                                'class_abundance': percent_abundance,
                                'class_count': peaks_count_percentile
                                }
                        
                        for atom in formula_dict.keys():
                        
                           dict_result[atom] = formula_dict.get(atom)

                    dict_data_list.append(dict_result)

                else:

                    if not include_unassigned: continue

                    dict_result = {'mz':  ms_peak._mz_exp,
                                'calibrated_mz': ms_peak.mz_exp,
                                'abundance': ms_peak.abundance,
                                'resolving_power': ms_peak.resolving_power,
                                'sn':  ms_peak.signal_to_noise,
                                'ion_charge': ms_peak.ion_charge,
                                'class': classe,
                                'class_abundance': percent_abundance,
                                'class_count': percent_abundance
                                }
                
                    dict_data_list.append(dict_result)                

        columns = columns_labels + self.all_identified_atoms

        return DataFrame(dict_data_list, columns=columns)
     
    def plot_ms_assigned_unassigned(self, assigned_color= 'b', unassigned_color = 'r'):
        
        mz_assigned = self.mz_exp_assigned()
        abundance_assigned = self.abundance_assigned()
    
        mz_not_assigned = self.mz_exp(Labels.unassigned)
        abundance_not_assigned = self.abundance(Labels.unassigned)
        
        ax = plt.gca()

        for plot_obj in ax.stem(mz_assigned,abundance_assigned, linefmt='-',  markerfmt=" ", use_line_collection =True, label="Assigned"):
        
            plt.setp(plot_obj, 'color', assigned_color, 'linewidth', 2)
        
        for plot_obj in ax.stem(mz_not_assigned, abundance_not_assigned, linefmt='-', markerfmt=" ",  use_line_collection =True,label="Unassigned"):
        
            plt.setp(plot_obj, 'color', unassigned_color, 'linewidth', 2)
        
        ax.set_xlabel("$\t{m/z}$", fontsize=12)
        ax.set_ylabel('Abundance', fontsize=12)
        ax.tick_params(axis='both', which='major', labelsize=12)

        ax.axes.spines['top'].set_visible(False)
        ax.axes.spines['right'].set_visible(False)

        ax.get_yaxis().set_visible(False)
        ax.spines['left'].set_visible(False)
        plt.legend()
        return ax    

    def plot_mz_error(self, color= 'g'):
        
        ax = plt.gca()

        mz_assigned = self.mz_exp_all()
        mz_error= self.mz_error_all()
        
        ax.scatter( mz_assigned, mz_error, c=color)
        
        ax.set_xlabel("$\t{m/z}$", fontsize=12)
        ax.set_ylabel('Error (ppm)', fontsize=12)
        ax.tick_params(axis='both', which='major', labelsize=12)

        ax.axes.spines['top'].set_visible(True)
        ax.axes.spines['right'].set_visible(True)

        ax.get_yaxis().set_visible(True)
        ax.spines['left'].set_visible(True)

        ax.set_xlim(self.min_max_mz)
        ax.set_ylim(self.min_ppm_error , self.max_ppm_error)
    
        return ax

    def plot_mz_error_class(self, classe, color= 'g'):
        
        if classe != Labels.unassigned:
            ax = plt.gca()
            
            abun_perc = self.abundance_count_percentile(classe)
            mz_assigned = self.mz_exp(classe)
            mz_error= self.mz_error(classe)
            
            ax.scatter( mz_assigned, mz_error, c=color)
            
            title = "%s, %.2f %%" % (classe, abun_perc)
            ax.set_title(title)
            ax.set_xlabel("$\t{m/z}$", fontsize=12)
            ax.set_ylabel('Error (ppm)', fontsize=12)
            ax.tick_params(axis='both', which='major', labelsize=12)

            ax.axes.spines['top'].set_visible(True)
            ax.axes.spines['right'].set_visible(True)

            ax.get_yaxis().set_visible(True)
            ax.spines['left'].set_visible(True)

            ax.set_xlim(self.min_max_mz)
            ax.set_ylim(self.min_ppm_error , self.max_ppm_error)
        
            return ax   
            
    def plot_ms_class(self, classe, color= 'g'):
        
        if classe != Labels.unassigned:
            ax = plt.gca()
            
            abun_perc = self.abundance_count_percentile(classe)
            mz_assigned = self.mz_exp(classe)
            abundance_assigned= self.abundance(classe)

            for plot_obj in ax.stem( mz_assigned, abundance_assigned, linefmt='-',  markerfmt=" ", use_line_collection =True):
            
                plt.setp(plot_obj, 'color', color, 'linewidth', 2)
            
            title = "%s, %.2f %%" % (classe, abun_perc)
            ax.set_title(title)
            ax.set_xlabel("$\t{m/z}$", fontsize=12)
            ax.set_ylabel('Abundance', fontsize=12)
            ax.tick_params(axis='both', which='major', labelsize=12)

            ax.axes.spines['top'].set_visible(False)
            ax.axes.spines['right'].set_visible(False)

            ax.get_yaxis().set_visible(False)
            ax.spines['left'].set_visible(False)

            ax.set_xlim(self.min_max_mz)
            ax.set_ylim(self.min_max_abundance)
        
            return ax

    def plot_van_krevelen(self, classe, max_hc=2.5, max_oc=2, ticks_number=5, color="jet"):
        
        if classe != Labels.unassigned:

            # get data 
            abun_perc = self.abundance_count_percentile(classe)
            hc = self.atoms_ratio(classe, "H", "C") 
            oc = self.atoms_ratio(classe, "O", "C") 
            abundance = self.abundance(classe)
            
            #plot data
            ax = plt.gca()

            ax.scatter(oc, hc, c=abundance, alpha=0.5, cmap=color)

            #ax.scatter(carbon_number, dbe, c=color, alpha=0.5)
            
            title = "%s, %.2f %%" % (classe, abun_perc)
            ax.set_title(title)
            ax.set_xlabel("O/C", fontsize=16)
            ax.set_ylabel('H/C', fontsize=16)
            ax.tick_params(axis='both', which='major', labelsize=18)
            ax.set_xticks(linspace(0, max_oc, ticks_number, endpoint=True))
            ax.set_yticks(linspace(0, max_hc, ticks_number, endpoint=True))

            # returns matplot axes obj and the class percentile of the relative abundance 
            
            return ax, abun_perc 
            
    def plot_dbe_vs_carbon_number(self, classe, max_c=50, max_dbe=40, dbe_incr=5, c_incr=10, color="jet"):
        
        if classe != Labels.unassigned:

            # get data 
            abun_perc = self.abundance_count_percentile(classe)
            carbon_number = self.carbon_number(classe)
            dbe = self.dbe(classe)
            abundance = self.abundance(classe)
            
            #plot data
            ax = plt.gca()

            ax.scatter(carbon_number, dbe, c=abundance, alpha=0.5, cmap=color)

            #ax.scatter(carbon_number, dbe, c=color, alpha=0.5)
            
            title = "%s, %.2f %%" % (classe, abun_perc)
            ax.set_title(title)
            ax.set_xlabel("Carbon number", fontsize=16)
            ax.set_ylabel('DBE', fontsize=16)
            ax.tick_params(axis='both', which='major', labelsize=18)
            ax.set_xticks(range(0, max_c, c_incr))
            ax.set_yticks(range(0, max_dbe, dbe_incr))

            # returns matplot axes obj and the class percentile of the relative abundance 
            
            return ax, abun_perc 




class KendrickGrouping:

    def mz_odd_even_index_lists(self):
        # get odd and even indexes lists
        
        even_idx = [] 
        odd_idx = []
        
        for i, mspeak in enumerate(self.mspeaks):
            
            if mspeak.nominal_mz_exp % 2 == 0:
                even_idx.append(i)
            else:
                odd_idx.append(i)
        
        return even_idx, odd_idx 

    def calc_error(self, current, test):
        
        return ((current-test)/current)*1e6
  
    
    def populate_kendrick_index_dict_error(self, list_indexes, sort=True):
        
        def error():

            return  abs(current_kmd_reference - next_mspeak.kmd)

        already_found = []
        
        all_results = []
        
        for i in list_indexes:
            
            result_indexes = []
            
            mspeak = self.mspeaks[i]    

            current_kmd_reference = mspeak.kmd
            for j in list_indexes :
                
                if j not in already_found and j != i:
                    
                    next_mspeak = self.mspeaks[j]

                    if  error() <= 0.001:
                        
                        result_indexes.append(j)    
                        already_found.append(j)

                        current_kmd_reference = next_mspeak.kmd
            
            if result_indexes and len(result_indexes) > 3:
                
                already_found.append(i)
                
                result_indexes.insert(0,i)
                
                all_results.append(result_indexes)        
            else:
                
                for w in result_indexes:

                    already_found.remove(w)        

        kendrick_group_index = { i : indexes_list for i, indexes_list in enumerate(all_results) }

       
            #return dictionary with the keys sorted by sum of the abundances
        if sort:
            
            return dict(sorted(kendrick_group_index.items(), key = lambda it: sum([self.mspeaks[i].abundance for i in it[1]]), reverse=False ))
        
        else:
            
            return kendrick_group_index
        
    def populate_kendrick_index_dict_rounding(self, list_indexes, sort=True):

        #self.test(list_indexes)
        #breakpoint()
        kendrick_group_index = {}
        
        for i in list_indexes:
            
            mspeak = self.mspeaks[i]
            
            group = round(mspeak.kmd * 100)
            
            if group not in kendrick_group_index:
                
                kendrick_group_index[group] = [i]

            else: 
                
                last_index = kendrick_group_index[group][-1]
                
                print(abs(mspeak.kmd - self.mspeaks[last_index].kmd ))
                
                if abs(mspeak.kmd - self.mspeaks[last_index].kmd ) < 0.001:

                    kendrick_group_index[group].append(i)
                
               


            #return dictionary with the keys sorted by sum of the abundances
        if sort:
            return dict(sorted(kendrick_group_index.items(), key = lambda it: sum([self.mspeaks[i].abundance for i in it[1]]), reverse=True ))
        
        else:
            return kendrick_group_index

    def sort_abundance_kendrick_dict(self, even_kendrick_group_index, odd_kendrick_group_index):
    
        all_even_indexes = [i for v in even_kendrick_group_index.values() for i in v]
        
        all_odd_indexes = [i for v in odd_kendrick_group_index.values() for i in v]

        sum_even = sum([self.mspeaks[i].abundance for i in all_even_indexes])
        
        sum_odd = sum([self.mspeaks[i].abundance for i in all_odd_indexes])
        
        if sum_even >= sum_odd:
            
            even_kendrick_group_index.update(odd_kendrick_group_index)
            
            return even_kendrick_group_index
        
        else: 
            
            odd_kendrick_group_index.update(even_kendrick_group_index)
            
            return odd_kendrick_group_index  
        
    
    def kendrick_groups_indexes(self, sort=True):
        
        #return dictionary with the kmd as keys and the correspondents peaks indexes
        even_idx, odd_idx = self.mz_odd_even_index_lists()
        
        even_kendrick_group_index = self.populate_kendrick_index_dict_error(even_idx, sort=sort)
        
        odd_kendrick_group_index = self.populate_kendrick_index_dict_error(odd_idx, sort=sort)
        
        return self.sort_abundance_kendrick_dict(even_kendrick_group_index, odd_kendrick_group_index)


'''
Created on Jun 14, 2019

@author: eber373
'''

__author__ = "Yuri E. Corilo"
__date__ = "Jun 14, 2019"

class LC_Calculations:
    
    '''
    classdocs
    '''
    
    def find_nearest_scan(self, rt):

        from numpy import abs as absolute
        from numpy import array

        array_rt = array(self.retention_time)

        scan_index = (absolute(array_rt - rt)).argmin()

        real_scan = self.scans_number[scan_index]

        return real_scan + 1
    




__author__ = "Yuri E. Corilo"
__date__ = "Jun 12, 2019"

from copy import deepcopy
from corems.ms_peak.calc.MSPeakCalc import MSPeakCalculation

class _MSPeak(MSPeakCalculation):
    '''
    classdocs
    '''
    def __init__(self, ion_charge, mz_exp, abundance, resolving_power, 
                    signal_to_noise, massspec_indexes, index, ms_parent=None, exp_freq=None):

        # needed to create the object
        self.ion_charge = int(ion_charge)
        self._mz_exp = float(mz_exp)
        self.mass = float(mz_exp) / float(ion_charge)
        self.abundance = float(abundance)
        self.resolving_power = float(resolving_power)
        self.signal_to_noise = float(signal_to_noise)
        # profile indexes
        self.start_index = int(massspec_indexes[0]) 
        self.apex_index = int(massspec_indexes[1])
        self.final_index = int(massspec_indexes[2]) 
        #centroid index
        self.index = int(index)
        # parent mass spectrum obj instance
        self._ms_parent = ms_parent
        
        # updated after mass error prediction'
        self.predicted_std = None
        # updated after calibration'
        self.mz_cal = None
        # updated individual calculation'
        self.baseline_noise = None
        
        if exp_freq:
            self.freq_exp = float(exp_freq)

        if self._ms_parent is not None:
            kendrick_dict_base = self._ms_parent.mspeaks_settings.kendrick_base
        else:
            kendrick_dict_base = {'C':1, 'H':2}
        self._kdm, self._kendrick_mass, self._nominal_km = self._calc_kdm(
            kendrick_dict_base)
 
        'updated after molecular formula ID'

        self.molecular_formulas = []
        self._confidence_score = None
        # placeholder for found isotopologues index 
        self.isotopologue_indexes = []
        # placeholder for found isotopologues molecular formula obj
        self.found_isotopologues = {}

    def __len__(self):
        
        return len(self.molecular_formulas)
        
    def __setitem__(self, position, molecular_formula_obj):
        
        self.molecular_formulas[position] = molecular_formula_obj

    def __getitem__(self, position):
        
        return self.molecular_formulas[position]

    def change_kendrick_base(self, kendrick_dict_base):
        '''kendrick_dict_base = {"C": 1, "H": 2}'''
        self._kdm, self._kendrick_mass, self._nominal_km = self._calc_kdm(
            kendrick_dict_base)

    def add_molecular_formula(self, molecular_formula_obj):
        
        # freeze state
        molecular_formula_obj._mspeak_parent = self
        
        #new_mol_formula = deepcopy(molecular_formula_obj)
        # add link mass spectrum obj instance
        
        #new_mol_formula.mspeak_parent = self

        self.molecular_formulas.append(molecular_formula_obj)

        return molecular_formula_obj

    def remove_molecular_formula(self, mf_obj):
        
        self.molecular_formulas.remove(mf_obj)

    def clear_molecular_formulas(self):
        
        self.molecular_formulas= []
    
    @property
    def mz_exp(self):
        if self.mz_cal:
            return self.mz_cal
        else:
            return self._mz_exp
    
    @mz_exp.setter
    def mz_exp(self, mz_exp):
        self._mz_exp = mz_exp

    @property
    def area(self): return self.calc_area()

    @property
    def nominal_mz_exp(self): return int(self.mz_exp)

    @property
    def kmd(self): return self._kdm

    @property
    def kendrick_mass(self): return self._kendrick_mass

    @property
    def knm(self): return self._nominal_km
    
    @property
    def is_assigned(self):

        return bool(self.molecular_formulas)
    
    def plot_simulation(self, sim_type="lorentz", ax=None, color="green",
                            oversample_multiplier=1, delta_rp = 0, mz_overlay=1):

                 
        if self._ms_parent:
        
            import matplotlib.pyplot as plt
            
            x, y = eval("self."+sim_type+"(oversample_multiplier="+str(oversample_multiplier)+", delta_rp="+str(delta_rp)+", mz_overlay="+str(mz_overlay)+")")
            
            if ax is None:
                    ax = plt.gca()
            ax.plot(x, y, color=color, label="Simulation")
            ax.set(xlabel='m/z', ylabel='abundance')
            
            plt.legend()
            return ax
           
    def plot(self, ax=None, color="black"): #pragma: no cover
        
        if self._ms_parent:
            
            import matplotlib.pyplot as plt

            if ax is None:
                ax = plt.gca()
            x = self._ms_parent.mz_exp_profile[self.start_index: self.final_index]
            y =  self._ms_parent.abundance_profile[self.start_index: self.final_index]
            
            ax.plot(x, y, color=color, label="Data")
            ax.set(xlabel='m/z', ylabel='abundance')
        
            #plt.legend()
        
            return ax
        
        else:
            print("Isolated Peak Object")

    @property
    def best_molecular_formula_candidate(self):
        
        if self._ms_parent.molecular_search_settings.score_method == "N_S_P_lowest_error":
            return self.cia_score_N_S_P_error()
        
        elif self._ms_parent.molecular_search_settings.score_method == "S_P_lowest_error":
            return self.cia_score_S_P_error()

        elif self._ms_parent.molecular_search_settings.score_method == "lowest_error":
            return self.molecular_formula_lowest_error()    
        
        elif self._ms_parent.molecular_search_settings.score_method == "air_filter_error":
            return self.molecular_formula_air_filter()    

        elif self._ms_parent.molecular_search_settings.score_method == "water_filter_error":
            return self.molecular_formula_water_filter()    

        elif self._ms_parent.molecular_search_settings.score_method == "earth_filter_error":
            return self.molecular_formula_earth_filter()   

        elif self._ms_parent.molecular_search_settings.score_method == "prob_score":
            return self.molecular_formula_highest_prob_score()
        else:
            
            raise TypeError("Unknown score method selected: % s, \
                            Please check score_method at \
                            encapsulation.settings.molecular_id.MolecularIDSettings.MolecularFormulaSearchSettings", 
                            self._ms_parent.parameters.molecular_search.score_method)    

class ICRMassPeak(_MSPeak):

    def __init__(self, *args, ms_parent=None, exp_freq=None):

        super().__init__(*args,exp_freq=exp_freq, ms_parent=ms_parent)

    def resolving_power_calc(self, B, T):
        
        '''
        low pressure limits, 
        T: float 
            transient time
        B: float
            Magnetic Filed Strength (Tesla)    
        
        reference
        Marshall et al. (Mass Spectrom Rev. 1998 Jan-Feb;17(1):1-35.)
        DOI: 10.1002/(SICI)1098-2787(1998)17:1<1::AID-MAS1>3.0.CO;2-K
        '''
        return (1.274e7 * self.ion_charge * B * T)/ (self.mz_exp*self.ion_charge)

    def set_calc_resolving_power(self, B, T):

        self.resolving_power = self.resolving_power_calc(B, T) 
        
class TOFMassPeak(_MSPeak):

    def __init__(self, *args, exp_freq=None):

        super().__init__(*args,exp_freq=exp_freq)

    def set_calc_resolving_power(self):
        return 0

class OrbiMassPeak(_MSPeak):

    def __init__(self, *args, exp_freq=None):

        super().__init__(*args,exp_freq=exp_freq)

    def set_calc_resolving_power(self):
        return 0       




__author__ = "Yuri E. Corilo"
__date__ = "Nov 11, 2019"


from io import StringIO, BytesIO
from pathlib import Path
from copy import deepcopy

from pandas import read_csv, read_pickle, read_excel
import chardet
from s3path import S3Path

from corems.encapsulation.factory.processingSetting import DataInputSetting
from corems.encapsulation.factory.parameters import default_parameters
from corems.encapsulation.constant import Labels
from corems.encapsulation.input.parameter_from_json import load_and_set_parameters_class, load_and_set_parameters_ms

class MassListBaseClass:
    '''
    # The MassListBaseClass object reads mass list data types and returns the mass spectrum obj

    # Parameters
    ----------
    ## file_location : Path or S3Path
        file_location is full data path
    ## * delimiter: str
        Delimeter to read text based files ("," "\t", " ", "  ", etc)
    ## **data_type : str
        The keyword argument data_type is used to determine what type of file to read:
        pandas(pickle), txt(.csv, txt, asci, etc) or HDF5
    ## **isCentroid : bool
        The keyword argument isCentroid is used to determine the mass spectrum data structure:
        if set to False will assume profile mode and will attempt to peak pick
    ## **header_lines : int
        Line count to skip: should include the columns labels line
     ## **isThermoProfile : bool
        The keyword argument is_thermo_profile is used to change the number of expected columns to only m/z and intensity
        S/N and RP will be calculated based on the data. i.e signal to noise might not be accurate because the lack of noise data 
    # Attributes
    ----------
    ## _expected_columns : set
       The file has to have a least the values inside the set.

    For label translation see:
                corems.encapsulation.settings.input.InputSetting 
        or add your labels to the SettingsCoreMS.json file and parse the settings

    '''

    def __init__(self, file_location, isCentroid=True, analyzer='Unknown', instrument_label='Unknown',
                 sample_name=None, header_lines=0, isThermoProfile=False):

        self.file_location = Path(file_location) if isinstance(file_location, str) else file_location
        
        if not self.file_location.exists():
            raise FileExistsError("File does not exist: %s" % file_location)

        # (newline="\n")

        self.header_lines = header_lines

        if isThermoProfile:
            
            self._expected_columns = {Labels.mz, Labels.abundance}

        else:

            self._expected_columns = {Labels.mz, Labels.abundance, Labels.s2n, Labels.rp}


        self._delimiter = None

        self.isCentroid = isCentroid

        self._data_type = None

        self.analyzer = analyzer

        self.instrument_label = instrument_label

        self.sample_name = sample_name

        self._parameters = deepcopy(DataInputSetting())

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, instance_DataInputSetting):
        self._parameters = instance_DataInputSetting

    def set_parameter_from_json(self, parameters_path):
        self._parameters = load_and_set_parameters_class(
            'DataInput', self.parameters, parameters_path=parameters_path)

    @property
    def data_type(self):
        return self._data_type

    @data_type.setter
    def data_type(self, data_type):
        self._data_type = data_type

    @property
    def delimiter(self):
        return self._delimiter

    @delimiter.setter
    def delimiter(self, delimiter):
        self._delimiter = delimiter

    def encoding_detector(self, file_location):
        with file_location.open('rb') as rawdata:
            result = chardet.detect(rawdata.read(10000))
        return result['encoding']

    def set_data_type(self):

        if self.file_location.suffix == '.csv':

            self.data_type = 'txt'
            self.delimiter = ','

        elif self.file_location.suffix == '.txt':

            self.data_type = 'txt'
            self.delimiter = '\t'

        elif self.file_location.suffix == '.xlsx':

            self.data_type = 'excel'

        elif self.file_location.suffix == '.ascii':

            self.data_type = 'txt'
            self.delimiter = '  '

        elif self.file_location.suffix == '.pkl':

            self.data_type = 'dataframe'
        else:
            raise TypeError(
                "Data type could not be automatically recognized for %s; please set data type and delimiter manually." % self.file_location.name)

    def get_dataframe(self):

        if not self.data_type or not self.delimiter:

            self.set_data_type()

        if isinstance(self.file_location, S3Path):
            # data = self.file_location.open('rb').read()
            data = BytesIO(self.file_location.open('rb').read())
        
        else:
            data = self.file_location

        if self.data_type == 'txt':
            
            dataframe = read_csv(data,  skiprows= self.header_lines, delimiter=self.delimiter,
                                 encoding=self.encoding_detector(self.file_location), engine='python')

        elif self.data_type == 'dataframe':

            dataframe = read_pickle(data)

        elif self.data_type == 'excel':

            dataframe = read_excel(data)

        else:

            raise TypeError('Data type %s is not supported' % self.data_type)

        return dataframe

    def load_settings(self, mass_spec_obj, output_parameters):

        import json
        import warnings

        settings_file_path = self.file_location.with_suffix('.json')

        if settings_file_path.exists():

            self._parameters = load_and_set_parameters_class(
                'DataInput', self._parameters, parameters_path=settings_file_path)

            load_and_set_parameters_ms(
                mass_spec_obj, parameters_path=settings_file_path)

        else:

            warnings.warn(
                "auto settings loading is enabled but could not locate the file:  %s. Please load the settings manually" % settings_file_path)

        # TODO this will load the setting from SettingCoreMS.json
        # coreMSHFD5 overrides this function to import the attrs stored in the h5 file
        #loaded_settings = {}
        #loaded_settings['MoleculaSearch'] = self.get_scan_group_attr_data(scan_index,  time_index, 'MoleculaSearchSetting')
        #loaded_settings['MassSpecPeak'] = self.get_scan_group_attr_data(scan_index,  time_index, 'MassSpecPeakSetting')

        #loaded_settings['MassSpectrum'] = self.get_scan_group_attr_data(scan_index, time_index, 'MassSpectrumSetting')
        #loaded_settings['Transient'] = self.get_scan_group_attr_data(scan_index, time_index, 'TransientSetting')

    def get_output_parameters(self, polarity, scan_index=0):

        # TODO pull attrs from json settings file in load_settings function MassSpecAttrs group and analyzer, instrument_label and sample_name
        from copy import deepcopy

        output_parameters = default_parameters(self.file_location)

        if self.isCentroid:
            output_parameters['label'] = Labels.corems_centroid
        else:
            output_parameters['label'] = Labels.bruker_profile

        output_parameters['analyzer'] = self.analyzer

        output_parameters['instrument_label'] = self.instrument_label

        output_parameters['sample_name'] = self.sample_name

        output_parameters["Aterm"] = None

        output_parameters["Bterm"] = None

        output_parameters["Cterm"] = None

        output_parameters["polarity"] = polarity

        '''scan_number and rt will be need to lc ms'''

        output_parameters["mobility_scan"] = 0

        output_parameters["mobility_rt"] = 0

        output_parameters["scan_number"] = scan_index

        output_parameters["rt"] = 0

        return output_parameters

    def clean_data_frame(self, dataframe):

        # print(dataframe.head())

        for column_name in dataframe.columns:

            expected_column_name = self.parameters.header_translate.get(
                column_name)
            if expected_column_name not in self._expected_columns:

                #dataframe = dataframe.drop(column_name, axis=1)
                del dataframe[column_name]
        # return dataframe

    def check_columns(self, header_labels):

        # print(header_labels)
        #print( self.parameters.header_translate.keys())
        #inverted_name_dict = {value: key for key, value in self.parameters.header_translate.items()}

        # print(inverted_name_dict)

        found_label = set()

        for label in header_labels:

            if not label in self._expected_columns:

                user_column_name = self.parameters.header_translate.get(label)

                if user_column_name in self._expected_columns:

                    found_label.add(user_column_name)

            else:
                found_label.add(label)

        not_found = self._expected_columns - found_label

        if len(not_found) > 0:

            raise Exception(
                "Please make sure to include the columns %s" % ', '.join(not_found))




from io import BytesIO

import h5py
from s3path import S3Path
from corems.mass_spectrum.input.baseClass import MassListBaseClass
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecProfile
from corems.encapsulation.constant import Labels
from corems.encapsulation.factory.parameters import default_parameters

class ReadHDF_BoosterMassSpectrum(MassListBaseClass):
    
    '''
    The MassSpectra object contains lots of MassSpectrum

    Parameters
    ----------
    arg : str
        The arg is used for ...
    *args
        The variable arguments are used for ...
    **kwargs
        The keyword arguments are used for ...

    Attributes
    ----------
    arg : str
        This is where we store arg,
    '''

    def __init__(self, file_location, isCentroid=False):
        
        self.polarity = self.get_polarity(file_location)

        super().__init__(file_location, isCentroid=False)
        
    def get_data_profile(self, mz, abundance, auto_process, auto_noise):

        data_dict = {Labels.mz: mz, Labels.abundance: abundance}
    
        output_parameters = self.get_output_parameters()
            
        return MassSpecProfile(data_dict, output_parameters, auto_process=auto_process, auto_noise=auto_noise)
    
    def get_attr_data(self, scan, attr_srt):

        return self.h5pydata[self.scans[scan]].attrs[attr_srt]

    def get_polarity(self, file_location):

        if isinstance(file_location, S3Path):
            data = BytesIO(file_location.open('rb').read())
        else:
            data = file_location
        
        self.h5pydata = h5py.File(data, 'r')

        self.scans = list(self.h5pydata.keys())
        
        polarity = self.get_attr_data(0,'r_h_polarity')
        
        if polarity == 'negative scan':
            
            return -1
        else:
            return +1    

    def get_mass_spectrum(self, auto_process=True, auto_noise=True):
        
        # only one mass spectrum
        if len(self.scans) == 1:
            
            booster_data = self.h5pydata[self.scans[0]]
            
            if self.isCentroid:
                
                raise NotImplementedError
                #return numpyArray.ms_from_array_centroid(mz, abun, rp, snt, self.file_location.stem, polarity=self.polarity,  auto_process=True, auto_noise=auto_noise)
                
            else:
                
                mz = booster_data[0]
                abun = booster_data[1]
                
                return self.get_data_profile(mz, abun, auto_process, auto_noise)

    def get_output_parameters(self):
        
        d_params = default_parameters(self.file_location)
        
        d_params["polarity"] = self.polarity
        
        d_params["filename_path"] = self.file_location
        
        d_params["mobility_scan"] = 0
        
        d_params["mobility_rt"] = 0
        
        d_params["scan_number"] = 0
        
        d_params["rt"] = self.get_attr_data(0, 'r_h_start_time')

        d_params['label'] = Labels.booster_profile

        d_params["Aterm"] = self.get_attr_data(0, 'r_cparams')[0]

        d_params["Bterm"] = self.get_attr_data(0, 'r_cparams')[1]

        return d_params


    
    
    


__author__ = 'Yuri E. Corilo'
__date__ = "03/31/2020"

from threading import Thread
from pandas import DataFrame
from numpy import hstack, inf, isnan, where, array
from tqdm import tqdm

class MassErrorPrediction(Thread):
    
    def __init__(self, mass_spectrum, mz_overlay=10, rp_increments=10000, base_line_target=0.01, max_interation=1000, interpolation='linear'):
        
        Thread.__init__(self)
        
        self.mass_spectrum_obj = mass_spectrum

        self.mz_overlay = mz_overlay

        self.rp_increments = 10000

        self.base_line_target = 0.01 

        self.max_interation = 1000

        self.df = None

        self.interpolation = interpolation
    
    def run(self):
            
        self.df = self.calc_error_dist()

    def get_results(self):

        if not self.df:
            self.run()

        return self.df

    def calc_error_dist(self):
        
        results_list = []
        
        indexes_without_results = list(range(len(self.mass_spectrum_obj)))
        # loop trough mass spectrum

        for peak_obj_idx, peak_obj in enumerate(tqdm(self.mass_spectrum_obj)):
            
            # access ms peaks triplets ( peak_obj_idx -1, peak_obj_idx, and peak_obj_idx + 1)
            # check lower and upper boundaries to not excesses mass spectrum range
            
            if  peak_obj_idx != 0 and peak_obj_idx != len(self.mass_spectrum_obj)-1:
                
                # current peak_obj initialted in the loop expression
                # geting the peak on the left (previous_peak_obj) and the one in the right position (next_peak_obj)
                next_peak_obj = self.mass_spectrum_obj[peak_obj_idx + 1]
                previous_peak_obj = self.mass_spectrum_obj[peak_obj_idx - 1]
                
                # check mz range defined in max_mz variable and check if peaks have same nominal mz
                # keeping same mz for better plotting representation only, remove it for production
                if  peak_obj.nominal_mz_exp == next_peak_obj.nominal_mz_exp and peak_obj.nominal_mz_exp == previous_peak_obj.nominal_mz_exp:
                    
                    #simulate peak shape
                    sim_mz, sim_abun = peak_obj.gaussian(mz_overlay=self.mz_overlay)
                    #update_plot(sim_mz,sim_abun, 0.5)
                    
                    #simulate peak shape
                    next_sim_mz, next_sim_abun = next_peak_obj.gaussian(mz_overlay=self.mz_overlay)
                    #update_plot(next_sim_mz, next_sim_abun, 0.5)
                    
                    
                    #simulate peak shape
                    previous_sim_mz, previous_sim_abun = previous_peak_obj.gaussian(mz_overlay=self.mz_overlay)
                    #update_plot(previous_sim_mz,  previous_sim_abun, 0.5)
                    
                    sim_mz_domain,  summed_peaks_abun = self.sum_data( ((previous_sim_mz,previous_sim_abun),  (sim_mz,sim_abun), (next_sim_mz, next_sim_abun)) )
                    #update_plot(sim_mz_domain,summed_peaks_abun, 0.5)
                    
                    #sum simulated abundances 
                    #summed_peaks_abun = (sim_abun + next_sim_abun + previous_sim_abun) 
                    
                    #normalize abundances to 0-1
                    #summed_peaks_abun = summed_peaks_abun/(max(summed_peaks_abun))

                    #find appexes location (mz) and magnitude
                    mz_centroid, abund_centroid = self.find_peak_apex(sim_mz_domain,summed_peaks_abun)    

                    #find valley location (mz_min_valley) and magnitude (abund_min_valley)
                    mz_min_valley, abund_min_valley = self.find_peak_valley(sim_mz_domain, summed_peaks_abun)  

                    # clear delta_rp (global implementation) and store choose resolving power increments   
                    delta_rp = self.rp_increments
                    
                    # used to limited number of iterations
                    i = 0
                    j = 0
                    
                    # TODO: fit peak shape and decide best fit #gaussian, lorentz and voigt 
                    #plot_triplets(mz_centroid,abund_centroid, mz_min_valley, abund_min_valley, sim_mz_domain, summed_peaks_abun )
                    if len(mz_centroid) == 2 :
                            
                        while len(mz_centroid) < 3 and i <= self.max_interation:
                            
                            previous_sim_mz, previous_sim_abun = previous_peak_obj.gaussian(delta_rp=delta_rp, mz_overlay=self.mz_overlay)
                            
                            sim_mz, sim_abun = peak_obj.gaussian(delta_rp=delta_rp, mz_overlay=self.mz_overlay)
                            
                            next_sim_mz, next_sim_abun = next_peak_obj.gaussian(delta_rp=delta_rp, mz_overlay=self.mz_overlay)

                            sim_mz_domain,  summed_peaks_abun = self.sum_data( ((previous_sim_mz,previous_sim_abun),  (sim_mz,sim_abun), (next_sim_mz, next_sim_abun)) )
                            
                            #update_plot(sim_mz_domain,  summed_peaks_abun, 0.01)

                            mz_centroid, abund_centroid = self.find_peak_apex(sim_mz_domain,summed_peaks_abun)    

                            delta_rp += self.rp_increments
                            
                            i += 1

                        mz_min_valley, abund_min_valley = self.find_peak_valley(sim_mz_domain, summed_peaks_abun)      

                    if len(mz_centroid) == 3 and len(abund_min_valley) == 2:
                        # increase all three peak resolving power until both valley magnitude is bellow the defined target
                        # calculate peak shapes with the needed resolving power to have a baseline resolution for all peaks
                        # calculate mass difference (ppm) between original centroid and the new simulated peak. 
                        
                        while  abund_min_valley[0] > self.base_line_target or abund_min_valley[1] > self.base_line_target and j <= self.max_interation:
                            
                            previous_sim_mz, previous_sim_abun = previous_peak_obj.gaussian(delta_rp=delta_rp, mz_overlay=self.mz_overlay)
                            
                            sim_mz, sim_abun = peak_obj.gaussian(delta_rp=delta_rp, mz_overlay=self.mz_overlay)
                            
                            next_sim_mz, next_sim_abun = next_peak_obj.gaussian(delta_rp=delta_rp, mz_overlay=self.mz_overlay)

                            sim_mz_domain,  summed_peaks_abun = self.sum_data( ((previous_sim_mz,previous_sim_abun),  (sim_mz,sim_abun), (next_sim_mz, next_sim_abun)) )
                            
                            #update_plot(sim_mz_domain,  summed_peaks_abun, 0.001)
                            
                            #summed_peaks_abun = (sim_abun + next_sim_abun + previous_sim_abun) 
                            
                            
                            #find appexes location (mz) and magnitude
                            mz_centroid, abund_centroid = self.find_peak_apex(sim_mz_domain,summed_peaks_abun)    
                            
                            #find valley location (mz_min_valley) and magnitude (abund_min_valley)
                            summed_peaks_abun = summed_peaks_abun/(summed_peaks_abun.max())
                            mz_min_valley, abund_min_valley = self.find_peak_valley(sim_mz_domain, summed_peaks_abun)  

                            if len(abund_min_valley) != 2:
                                break
                            
                            delta_rp += self.rp_increments
                            j += 1
                            
                            #plot_triplets(mz_centroid,abund_centroid, mz_min_valley, abund_min_valley, sim_mz_domain, summed_peaks_abun )
                        
                        
                        #plot_triplets(mz_centroid,abund_centroid, mz_min_valley, abund_min_valley, sim_mz_domain, summed_peaks_abun )

                        mass_shift_ppp = self.calc_error(mz_centroid[1], peak_obj.mz_exp, 1000000)
                        #delta_mz = mz_centroid[1] - peak_obj.mz_exp
                        height_shift_per = self.calc_error(abund_centroid[1], peak_obj.abundance, 100)
                        #excitation_amplitude = str(mass_spectrum_obj.filename.stem).split("ex")[1].split("pc")[0]
                        #ion_time = str(mass_spectrum_obj.filename.stem).split("0pt")[1].split("s")[0]
                        peak_obj.predicted_std = mass_shift_ppp
                        
                        results_list.append( {
                        "ms_index_position" : peak_obj_idx,
                        "predicted_std": mass_shift_ppp,
                        "mz_exp": peak_obj.mz_exp,
                        "nominal_mz_exp": peak_obj.nominal_mz_exp,
                        "predicted_mz": mz_centroid[1],
                        "s2n" : peak_obj.signal_to_noise,
                        "peak_height" : peak_obj.abundance,
                        "predicted_peak_height" : abund_centroid[1],
                        "peak_height_error" : height_shift_per,
                        "resolving_power" : peak_obj.resolving_power,
                        #"excitation_amplitude" : excitation_amplitude,
                        #"ion_time" : ion_time
                        })
                        
                        indexes_without_results.remove(peak_obj_idx)
                    #elif len(mz_centroid) == 3 and len(abund_min_valley) != 2:

        for peak_obj_idx in indexes_without_results:

            results_list.append( {
            "ms_index_position" : peak_obj_idx,
            "mz_exp": self.mass_spectrum_obj[peak_obj_idx].mz_exp,
            "nominal_mz_exp": self.mass_spectrum_obj[peak_obj_idx].nominal_mz_exp,
            "s2n" : self.mass_spectrum_obj[peak_obj_idx].signal_to_noise,
            "peak_height" : self.mass_spectrum_obj[peak_obj_idx].abundance,
            "resolving_power" : self.mass_spectrum_obj[peak_obj_idx].resolving_power,
            #"excitation_amplitude" : excitation_amplitude,
            #"ion_time" : ion_time
            } )

        df = DataFrame(results_list).sort_values("mz_exp")
        
        df.interpolate(method ='linear', limit_direction ='backward',  inplace=True)
        df.interpolate(method ='linear', limit_direction ='forward',  inplace=True)

        #TODO improve interpolation for missing data
        #f1 = interpolate.interp1d(x1, y1, kind='quadratic',fill_value="extrapolate")

        
        for peak_obj_idx in indexes_without_results:

            predicted_std = df.loc[peak_obj_idx].predicted_std
            
            self.mass_spectrum_obj[peak_obj_idx].predicted_std = predicted_std

        return df

    def sum_data(self, tuple_mz_abun_list):

        all_mz = {}

        for mz_list, abun_list in tuple_mz_abun_list:
            
            for index, mz in enumerate(mz_list):

                abundance = abun_list[index]

                if mz in all_mz:
                    all_mz[mz] = all_mz[mz] + abundance    
                else: 
                    all_mz[mz] = abundance
        
        mz_all = []
        abun_all = []

        for mz in sorted (all_mz) : 
            mz_all.append(mz)
            abun_all.append(all_mz[mz])

        return array(mz_all), array(abun_all)    

    def calc_error(self, mass_ref, mass_sim, factor):

        return ((mass_ref-mass_sim)/mass_ref)*factor

    def find_peak_apex(self, mass, abund):

        dy = abund[1:] - abund[:-1]

        '''replaces nan for infinity'''
        indices_nan = where(isnan(abund))[0]

        if indices_nan.size:
            
            abund[indices_nan] = inf
            dy[where(isnan(dy))[0]] = inf

        indexes = where((hstack((dy, 0)) < 0) & (hstack((0, dy)) > 0))[0]

        if indexes.size:
            
            return mass[indexes], abund[indexes]

    def find_peak_valley(self, mz, abund):

        dy = abund[1:] - abund[:-1]
        
        '''replaces nan for infinity'''
        indices_nan = where(isnan(abund))[0]
        
        if indices_nan.size:
            
            abund[indices_nan] = inf
            dy[where(isnan(dy))[0]] = inf
        
        indexes = where((hstack((dy, 0)) > 0) & (hstack((0, dy)) < 0))[0]

        return mz[indexes], abund[indexes]    